# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Main MCMC module."""

__all__ = [
    "distribution",
    "sampler",
    "gmrf",
    "mcmc",
    "model",
    "parameter",
]



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Collection of functions defining various MCMC samplers.

MCMCSampler is a superclass for all MCMC sampler types. This file contains several conjugate MCMC sampling algorithms,
which can be used for specific distribution combinations, and inherit directly from MCMCSampler.

metropolis_hastings.py contains another set of algorithms, all of Metropolis-Hastings type, where a parameter proposal
followed by an accept/reject step.

reversible_jump.py contains a generic implementation of the reversible jump algorithm.

"""

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from typing import Union

import numpy as np
from scipy import sparse
from scipy.stats import gamma, norm

from openmcmc import gmrf
from openmcmc.distribution.location_scale import Normal
from openmcmc.model import Model
from openmcmc.parameter import (
    Identity,
    MixtureParameterMatrix,
    MixtureParameterVector,
    ScaledMatrix,
)


@dataclass
class MCMCSampler(ABC):
    """Abstract base class for openMCMC sampling algorithms for a model parameter.

    Attributes:
        param (str): label of the parameter to be sampled.
        model (Model): sub-model of overall model, containing only distributions with some dependence on self.param.
        max_variable_size (Union[int, tuple]): (if required) maximum size for the variable. Only relevant in cases
            (e.g. RJMCMC) where variable dimension changes as a result of MCMC proposals.

    """

    param: str
    model: Model
    max_variable_size: Union[int, tuple, None] = None

    def __post_init__(self):
        """Extract the sub-model of distributions with some dependence on self.param."""
        self.model = self.model.conditional(self.param)

    @abstractmethod
    def sample(self, current_state: dict) -> dict:
        """Generate the next sample in the chain.

        Args:
            current_state (dict): dictionary containing current parameter values.

        Returns:
            (dict): state with the value of self.param updated to a new sample.

        """

    def init_store(self, current_state: dict, store: dict, n_iterations: int) -> dict:
        """Initialise the field in the MCMC storage dictionary for self.param.

        Args:
            current_state (dict): dictionary containing current parameter values.
            store (dict): dictionary to store all samples generated by the MCMC algorithm.
            n_iterations (int): total number of MCMC iterations to be run.

        Returns:
            (dict): storage dictionary updated with field for self.param.

        """
        if self.max_variable_size is None:
            store[self.param] = np.full(shape=(np.size(current_state[self.param]), n_iterations), fill_value=np.nan)
        elif isinstance(self.max_variable_size, tuple):
            store[self.param] = np.full(shape=self.max_variable_size + (n_iterations,), fill_value=np.nan)
        else:
            store[self.param] = np.full(shape=(self.max_variable_size, n_iterations), fill_value=np.nan)
        return store

    def store(self, current_state: dict, store: dict, iteration: int) -> dict:
        """Store the current state of the sampled variable in the MCMC storage dictionary.

        If self.parameter is not initialised in the MCMC state, then the function generates a random sample from the
        corresponding distribution in model.

        Args:
            current_state (dict): dictionary with current parameter values.
            store (dict): storage dictionary for MCMC samples.
            iteration (int): current MCMC iteration index.

        Returns:
            dict: storage dictionary updated with values from current iteration.

        """
        current_param = current_state[self.param]

        if self.max_variable_size is None:
            store[self.param][:, [iteration]] = current_param
        elif isinstance(self.max_variable_size, tuple):
            index_list = []
            for dim in range(current_param.ndim):
                index_list.append(np.arange(current_param.shape[dim], dtype=int))
            index_list.append(np.array([iteration]))

            store[self.param][np.ix_(*index_list)] = current_param.reshape(current_param.shape + (1,))
        else:
            store[self.param][range(current_param.size), [iteration]] = current_param.flatten()

        return store


@dataclass
class NormalNormal(MCMCSampler):
    """Normal-Normal conditional sampling (exploiting conjugacy).

    Sample from f(x|{b_k}) ~= [prod_k f(y_k|x)]f(x) where the components have the following form:
        - Likelihoods: f(y_k|x) ~ N(y_k | d_k + A_k*a, W_k^{-1})
        - Prior: f(x) ~ N(x |m, P^{-1})
    The following features are assumed:
        - There can be multiple likelihood/response distributions, but there is only one prior distribution.
        - The mean of each of the response distributions must have a linear dependence on self.param.
        - The prior Gaussian can be truncated (as long as the truncation points are constant), but the response
            Gaussians cannot.

    If the prior Gaussian is truncated (i.e. it has specified domain limits), then those same domain limits will be
    used when generating the conditional sample.

    Attributes:
        _is_response (dict): dictionary containing boolean indicators for whether self.param is the response of the
            distribution. If self._is_response[key] is True, then self.param is the response of the distribution.

    """

    def __post_init__(self):
        """Identify and extract the sub-model with a dependence on self.param.

        Also identify whether self.param is the response for each of the pair of Gaussian distributions.

        """
        super().__post_init__()
        self._is_response = {}
        for key in self.model.keys():
            self._is_response[key] = key == self.param

    def sample(self, current_state: dict) -> dict:
        """Generate a sample from a Gaussian-Gaussian conditional distribution.

        For a Gaussian-Gaussian conditional distribution, the parameters are as follows:
            Conditional precision:
                Q = P + sum_k [A_k'*W_k*A_k]
            Conditional mean:
                b = P*m + sum_k [A_k'*W_k*(y_k - d_k)]
                mu = Q^{-1} * b
        Where the parameters are as defined in the class docstring.

        If the supplied response parameter has a second dimension, these are interpreted as repeated draws from the same
        distribution, and are thus summed. The multiplication of the precision matrix by num_replicates is handled by
        the grad_log_p() function of the corresponding distribution.

        Args:
            current_state (dict): dictionary containing the current sampler state.

        Returns:
            (dict): state with updated value for self.param.

        """
        n_param = current_state[self.param].shape[0]
        Q = sparse.csc_matrix((n_param, n_param))
        b = np.zeros(shape=(n_param, 1))
        for key, dist in self.model.items():
            Q_rsp = dist.precision.predictor(current_state)
            if self._is_response[key]:
                Q += Q_rsp
                b += Q_rsp @ dist.mean.predictor(current_state)
            else:
                _, Q_dist = dist.grad_log_p(current_state, self.param)
                Q += Q_dist
                if isinstance(dist.mean, Identity):
                    b += Q_rsp @ np.sum(current_state[key], axis=1, keepdims=True)
                else:
                    predictor_exclude = dist.mean.predictor_conditional(current_state, term_to_exclude=self.param)
                    A = current_state[dist.mean.form[self.param]]
                    b += A.T @ Q_rsp @ (current_state[key] - predictor_exclude)

        dist_param = self.model[self.param]

        if dist_param.domain_response_lower is None and dist_param.domain_response_upper is None:
            current_state[self.param] = gmrf.sample_normal_canonical(b, Q)
        else:
            current_state[self.param] = gmrf.gibbs_canonical_truncated_normal(
                b,
                Q,
                x=current_state[self.param],
                lower=dist_param.domain_response_lower,
                upper=dist_param.domain_response_upper,
            )

        return current_state


@dataclass
class NormalGamma(MCMCSampler):
    """Normal-gamma conditional sampling (exploiting conjugacy).

    Assumes that self.param is the precision parameter for a Gaussian response distribution, and that it has a gamma
    prior distribution.

    Allows for the possibility that a single Gaussian response distribution might be associated with a number of
    different precision parameters, through use of a MixtureParameterMatrix precision parameters. These parameters
    are sampled in a loop within the sample function.

    This class samples from f(lam|y, a, b) ~ prod_k [f(y_k|lam)]f(lam|a,b) where
        - Likelihoods: f(y_k|lam) ~ N(mu_k, 1/lam)
        - Prior: f(lam|a, b) ~ G(a, b)
    Note that it is also possible to use a more complex (e.g. dense) precision matrix scaled by a single precision
    parameter, this does not fundamentally change the approach.

    Attributes:
        param (str): label of parameter name to be sampled
        normal_param (str): label of corresponding normal parameter
        model (Model): conditional model with distributions related to param only

    """

    def __post_init__(self):
        """Complete initialization of sampler.

        Identifies the gamma distribution and its conjugate normal distribution, and attaches copies to the sampler
        object.

        """
        super().__post_init__()

        nrm_prm = list(self.model.keys())
        nrm_prm.remove(self.param)
        self.normal_param = nrm_prm[0]

        precision = self.model[self.normal_param].precision

        if not isinstance(precision, (Identity, ScaledMatrix, MixtureParameterMatrix)):
            raise TypeError("precision must be either Identity, ScaledMatrix or MixtureParameterMatrix")

    def sample(self, current_state: dict) -> dict:
        """Generate a sample from a (series of) Gaussian-Gamma conditional distribution.

        The conditional distribution for an individual parameter is:
            G(a*, b*)
        where:
            a* = a_0 + n/2
            b* = b_0 + (y - y_hat)' * P * (y - y_hat)
        where n = [dimension of normal response], P = [un-scaled precision matrix].

        It is assumed that the precision parameter of the Gaussian distribution has a predictor_unscaled() method,
        which can be used to identify the subset of the full precision matrix which is dependent on the parameter being
        sampled: this is returned un-scaled by the precision parameter.

        Args:
            current_state (dict): dictionary containing the current sampler state.

        Returns:
            (dict): state with updated value for self.param.

        """
        precision = self.model[self.normal_param].precision
        mean = self.model[self.normal_param].mean
        y = current_state[self.model[self.normal_param].response]
        residual = y - mean.predictor(current_state)

        a = deepcopy(self.model[self.param].shape.predictor(current_state))
        b = deepcopy(self.model[self.param].rate.predictor(current_state))

        for k in range(current_state[self.param].shape[0]):
            precision_unscaled = precision.precision_unscaled(current_state, k)
            a[k] += np.sum(precision_unscaled.diagonal() > 0) / 2
            b[k] += (residual.T @ precision_unscaled @ residual).item() / 2
        no_warning_b = np.where(b == 0, np.inf, b)
        no_warning_scale = np.where(b == 0, np.inf, 1 / no_warning_b)
        current_state[self.param] = gamma.rvs(a, scale=no_warning_scale).reshape(current_state[self.param].shape)
        return current_state


@dataclass
class MixtureAllocation(MCMCSampler):
    """Conditional conjugate sampling of the allocation in a mixture distribution. Can be used with any kind of mixture distribution.

    This class samples from:
        f(z|y, lam, tht) ~ f(y|z, lam)f(z|tht)
    where:
        - Likelihood: f(y|z, lam); Gaussian distribution, with different characteristics depending on z.
        - Prior: f(z|tht); categorical distribution, with prior allocation probabilities tht.

    Attributes:
        response_param (str): name of the response parameter associated with the allocation parameter.

    """

    response_param: Union[str, None] = None

    def __post_init__(self):
        """Subset only model elements relevant for this sampler."""
        self.model = Model([self.model[self.param], self.model[self.response_param]])

        if not isinstance(self.model[self.response_param], Normal):
            raise TypeError("Mixture model currently only implemented for Normal case")

        if not isinstance(self.model[self.response_param].mean, MixtureParameterVector):
            raise TypeError("Mean must be of type MixtureParameterVector")

        if not isinstance(self.model[self.response_param].precision, MixtureParameterMatrix):
            raise TypeError("Mean must be of type MixtureParameterMatrix")

    def sample(self, current_state: dict) -> dict:
        """Generate sample of a parameter allocation given current state of the sampler.

        Computes the conditional allocation probability for each element of the response to each component of the
        mixture, then samples an allocation based on the probabilities.

        The conditional distribution is:
            Cat([gam_1, gam_2,..., gam_m])
        where:
            gam_k = f(y|z_k, lam) * tht_k / W
            W = sum_k [f(y|z_k, lam) * tht_k / Z]

        Args:
            current_state (dict): dictionary containing the current sampler state.

        Returns:
            (dict): state with updated value for self.param.

        """
        allocation_prior = self.model[self.param].prob.predictor(current_state)
        n_response = current_state[self.response_param].shape[0]
        component_mean = current_state[self.model[self.response_param].mean.param]
        component_precision = current_state[self.model[self.response_param].precision.param]

        allocation_prob = np.empty((n_response, allocation_prior.shape[1]))
        for k in range(allocation_prior.shape[1]):
            allocation_prob[:, [k]] = allocation_prior[:, [k]] * norm.pdf(
                current_state[self.response_param], loc=component_mean[k], scale=1 / np.sqrt(component_precision[k])
            )

        allocation_prob = allocation_prob / np.sum(allocation_prob, axis=1).reshape((allocation_prob.shape[0], 1))
        U = np.random.rand(n_response, 1)
        current_state[self.param] = np.atleast_2d(np.sum(U > np.cumsum(allocation_prob, axis=1), axis=1)).T

        return current_state



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Collection of possible parameter specifications for the distribution objects.

Example choices defined:

Identity:  f = x
LinearCombination:  f = X @ beta + Y @ gamma
LinearCombinationWithTransform: f = X @ exp(beta) + Y @ gamma
ScaledMatrix f = lam * P
MixtureParameterVector  f= X[I]
MixtureParameterMatrix  f= np.diag(lam[I])

"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Union

import numpy as np
from scipy import sparse


@dataclass
class Parameter(ABC):
    """Abstract base class for parameter."""

    @abstractmethod
    def predictor(self, state: dict) -> np.ndarray:
        """Create predictor from the state dictionary using the functional form defined in the specific subclass.

        Args:
            state (dict): dictionary object containing the current state information

        Returns:
            (np.ndarray): predictor vector

        """

    @abstractmethod
    def get_param_list(self) -> list:
        """Extract list of components from parameter specification.

        Returns:
            (list): parameter included as part of predictor

        """

    @abstractmethod
    def get_grad_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """

    @abstractmethod
    def grad(self, state: dict, param: str) -> np.ndarray:
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """


@dataclass
class Identity(Parameter):
    """Class specifying a simple predictor in a single term.

    Predictor has the functional form:
        f = x

    The gradient should only be used for scalar and vector inputs

    Args:
        form (str): string specifying the element of state which determines the parameter

    Attributes:
        form (str): string specifying the element of state which determines the parameter.

    """

    form: str

    def predictor(self, state: dict) -> np.ndarray:
        """Create predictor from the state dictionary using the functional form defined in the specific subclass.

        Args:
            state (dict): dictionary object containing the current state information

        Returns:
            (np.ndarray): predictor vector

        """
        return state[self.form]

    def get_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return [self.form]

    def get_grad_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return [self.form]

    def grad(self, state: dict, param: str) -> np.ndarray:
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """
        if state[self.form].shape[1] > 1:
            raise ValueError("Gradient in Identity should not be used for variables 2D and above.")
        p = state[self.form].size
        if param == self.form:
            grad = np.eye(p)
        else:
            grad = np.zeros(shape=(p, p))
        return grad


@dataclass
class LinearCombination(Parameter):
    """Class specifying linear combination form .

    This Parameter type is typically in the mean of a Normal distribution in a linear regression type case.

    Predictor has the form
        predictor  = sum_i (value[i] @ key[i])
    using the form dictionary input

    Attributes:
        form (dict): dict specifying the term and prefactor in the linear combination.
            example: {'beta': 'X', 'alpha': 'A'} produces linear combination X @ beta + A @ alpha.

    """

    form: dict

    def predictor(self, state: dict) -> np.ndarray:
        """Create predictor from the state dictionary using the functional form defined in the specific subclass.

        Args:
            state (dict): dictionary object containing the current state information

        Returns:
            (np.ndarray): predictor vector

        """
        return self.predictor_conditional(state)

    def predictor_conditional(self, state: dict, term_to_exclude: Union[str, list] = None) -> np.ndarray:
        """Extract predictor from the state dictionary using the functional form defined in the specific subclass excluding parameters.

        Used when estimating conditional distributions of those parameters.

        Args:
            state (dict): dictionary object containing the current state information
            term_to_exclude (Union[str, list]): terms to exclude from predictor

        Returns:
            (np.ndarray): predictor vector

        """
        if term_to_exclude is None:
            term_to_exclude = []

        if isinstance(term_to_exclude, str):
            term_to_exclude = [term_to_exclude]

        sum_terms = 0
        for prm, prefactor in self.form.items():
            if prm not in term_to_exclude:
                sum_terms += state[prefactor] @ state[prm]
        return sum_terms

    def get_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return list(self.form.keys()) + list(self.form.values())

    def get_grad_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return list(self.form.keys())

    def grad(self, state: dict, param: str) -> np.ndarray:
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """
        return state[self.form[param]].T


@dataclass
class LinearCombinationWithTransform(LinearCombination):
    """Linear combination of parameters from the state, with optional exponential transformation for the parameter elements.

    Currently, the only allowed transformation is the exponential transform.

    This Parameter type is typically in the mean of a Normal distribution and could be
    used to impose positivity of the parameters

    Predictor has the form
        predictor  = sum_i (value[i] @ transform(key[i]))
    using the form dictionary input

    Attributes:
        transform (dict): dict with logicals specifying whether exp(.) transform should
            be applied to parameter
            example: form={'beta': X}, transform={'beta': True} will produce X @ np.exp(beta)

    """

    transform: dict

    def predictor_conditional(self, state: dict, term_to_exclude: Union[str, list] = None) -> np.ndarray:
        """Extract predictor from the state dictionary using the functional form defined in the specific subclass excluding parameters.

        Used when estimating conditional distributions of those parameters.

        Args:
            state (dict): dictionary object containing the current state information
            term_to_exclude (list): terms to exclude from predictor

        Returns:
            (np.ndarray): predictor vector

        """
        if term_to_exclude is None:
            term_to_exclude = []

        if isinstance(term_to_exclude, str):
            term_to_exclude = [term_to_exclude]

        sum_terms = 0
        for prm, prefactor in self.form.items():
            if prm not in term_to_exclude:
                param = state[prm]
                if self.transform[prm]:
                    param = np.exp(param)
                sum_terms += state[prefactor] @ param
        return sum_terms

    def grad(self, state: dict, param: str) -> np.ndarray:
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """
        if self.transform[param]:
            if sparse.issparse(state[self.form[param]]):
                return state[self.form[param]].multiply(np.exp(state[param]).flatten()).T
            return np.exp(state[param]) * (state[self.form[param]].T)

        return state[self.form[param]].T


@dataclass
class ScaledMatrix(Parameter):
    """Defines parameter a scalar factor in front of a matrix.

    This is often used in case where we have a scalar variance in front of an unscaled precision matrix.
    Where we have a gamma distribution for the scalar parameter which wish to estimate

    Linear combinations have the form:
        predictor = scalar * matrix

    Attributes:
        matrix (str): variable name of the un-scaled matrix
        scalar (str): variable name of the scalar term

    """

    matrix: str
    scalar: str

    def predictor(self, state: dict) -> np.ndarray:
        """Create predictor from the state dictionary using the functional form defined in the specific subclass.

        Args:
            state (dict): dictionary object containing the current state information

        Returns:
            (np.ndarray): predictor vector

        """
        return float(state[self.scalar].item()) * state[self.matrix]

    def get_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return [self.scalar, self.matrix]

    def get_grad_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return [self.scalar]

    def grad(self, state: dict, param: str) -> np.ndarray:
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """
        return state[self.matrix]

    def precision_unscaled(self, state: dict, _) -> np.ndarray:
        """Return the precision matrix un-scaled by the scalar precision parameter.

        Args:
            state (dict): state dictionary
            _ (int): argument unused but matches with version in MixtureParameterMatrix where element is needed

        Returns:
            (np.ndarray): unscaled precision matrix

        """
        return state[self.matrix]


@dataclass
class MixtureParameter(Parameter, ABC):
    """Abstract Parameter class for a mixture distribution.

    Subclasses implemented for both:

    - vector-valued parameter (MixtureParameterVector)
    - diagonal matrix-valued parameter (MixtureParameterMatrix)
    where the elements of the vector or matrix diagonal are allocated based
    on the allocation parameter.

    """

    param: str
    allocation: str

    def get_element_match(self, state: dict, element_index: Union[int, np.ndarray]) -> np.ndarray:
        """Extract the parts of self.allocation which have given element number.

        used in the gradient function to pull out gradient for given element.

        Args:
            state (dict): state vector
            element_index (int, np.array): element index or set of integers

        Returns:
            (np.array(dtype=int)): element matches with 1 where there is a match and 0 where there isn't

        """
        if isinstance(element_index, np.ndarray) and element_index.size > 1:
            element_index = element_index.reshape((1, -1))

        return np.array(state[self.allocation] == element_index, dtype=int, ndmin=2)

    def get_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return [self.param, self.allocation]


@dataclass
class MixtureParameterVector(MixtureParameter):
    """Vector parameter: elements of the vector are obtained from sub-parameter 'param' according to the allocation.

    The allocation parameter defines a mapping between a R^m and R^n where typically m<=n and m is the
    number true underlying number of parameters in the model but due to the representation/algebra in
    other parts of the model this is expanded out to an n parameter model where the values of m are copied
    according to the index vector

    predictor = param [allocation]

    Attributes:
        param (str): name of underlying state component used to generate parameter.
        allocation (np.ndarray): name of allocation parameter within state dict.

    """

    def predictor(self, state: dict) -> np.ndarray:
        """Create predictor from the state dictionary using the functional form defined in the specific subclass.

        Args:
            state (dict): dictionary object containing the current state information

        Returns:
            (np.ndarray): predictor vector

        """
        return state[self.param][state[self.allocation].flatten()]

    def grad(self, state: dict, param: str):
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """
        element_index = np.arange(0, state[param].size)

        return self.get_element_match(state, element_index).astype(np.float64).T

    def get_grad_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return [self.param]


@dataclass
class MixtureParameterMatrix(MixtureParameter):
    """Diagonal matrix parameter: elements of the diagonal are obtained from sub-parameter 'param' according to the allocation index vector.

    The allocation parameter defines a mapping between a R^m and R^n where typically m<=n and m is the
    number true underlying number of parameters in the model but due to the representation/algebra in
    other parts of the model this is expanded out to an n parameter model where the values of m are copied
    according to the index vector

    predictor = np.diag( param [allocation] )

    Attributes:
        param (str): name of underlying state component used to generate parameter.
        allocation (np.ndarray): name of allocation parameter within state dict.

    """

    def predictor(self, state: dict) -> sparse.csc_matrix:
        """Create predictor from the state dictionary using the functional form defined in the specific subclass.

        Args:
            state (dict): dictionary object containing the current state information

        Returns:
            (sparse.csc_matrix): predictor vector

        """
        return sparse.diags(diagonals=state[self.param][state[self.allocation]].flatten(), offsets=0, format="csc")

    def grad(self, state: dict, param: str):
        """Compute gradient of single parameter.

        Args:
            state (dict): Dictionary object containing the current state information
            param (str): Compute derivatives WRT this variable

        Returns:
            (np.ndarray): [n_param x n_data] array, gradient with respect to param

        """
        raise TypeError("Not defined in this case")

    def get_grad_param_list(self) -> list:
        """Extract list of components from parameter specification that grad is defined for.

        Returns:
            (list): parameter that grad is defined for.

        """
        return []

    def precision_unscaled(self, state: dict, element_index: int) -> np.ndarray:
        """Return the precision matrix un-scaled by the scalar precision parameter.

        Args:
            state (dict): state dictionary
            element_index (int): index of element to subset

        Returns:
            (np.ndarray): unscaled precision matrix

        """
        return sparse.diags(diagonals=self.get_element_match(state, element_index).flatten(), offsets=0, format="csc")



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Collection of distributions for use with openMCMC code.

General assumptions about code functionality:
    - The first dimension of a parameter array is assumed to represent the dimensionality of the parameter vector; the
        second dimension is assumed to represent independent realizations of the parameter set. For example: an array
        with shape=(d, n) would be assumed to hold n replicates of a d-dimensional parameter vector.
    - self.response is a string containing the name of the response parameter for the distribution. For example, when
        self.response="y", all functions within the class will perform calculations using the value stored in
        state["y"].

"""

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from typing import Tuple, Union

import numpy as np
from scipy import sparse, stats

from openmcmc.parameter import Identity, LinearCombination, MixtureParameterVector


@dataclass
class Distribution(ABC):
    """Abstract superclass for handling distribution objects.

    Attributes:
        response (str): specifies the name of the response variable of the distribution.

    """

    response: str

    @abstractmethod
    def log_p(self, state: dict, by_observation: bool = False) -> Union[np.ndarray, float]:
        """Compute the log of the probability density (for current parameter settings).

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of responses; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p responses of
                the distribution separately. Defaults to False.

        Returns:
            (Union[np.ndarray, float]): POSITIVE log-density evaluated using the supplied state dictionary.

        """

    @abstractmethod
    def rvs(self, state: dict, n: int = 1) -> np.ndarray:
        """Generate random samples from the distribution.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (np.ndarray): random variables generated from distribution returned as p x n where p is the
                dimensionality of the response.

        """

    @property
    @abstractmethod
    def _dist_params(self) -> list:
        """Get list of parameter labels across all Parameter objects in distribution (EXCLUDING the response).

        Returns:
            (list): list of parameter labels.

        """

    @property
    def param_list(self) -> list:
        """Get list of all parameter labels in model (INCLUDING the response).

        Returns:
            (list): list of parameter labels

        """
        lst = [self.response] + self._dist_params
        return lst

    def grad_log_p(
        self, state: dict, param: str, hessian_required: bool = True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Generate vector of derivatives of the log-pdf with respect to a given parameter, and if required, also generate the Hessian.

        Function only defined for scalar- and vector-valued parameters param. If hessian_required=True, this function
        returns a tuple of (gradient, Hessian). If hessian_required=False, this function returns a np.ndarray (just
        the gradient of the log-density).

        As a default, the individual gradients are computed by finite-differencing the log_p function defined for the
        distribution. Where analytical forms for the gradient exist, these will be defined in distribution-specific
        subclasses.

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            hessian_required (bool): flag for whether the Hessian should be calculated and supplied as an output.

        Returns:
            (Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]): if hessian_required=True, then a tuple of (gradient,
                hessian) is returned. If hessian_required=False, then just a gradient vector is returned. The returned
                values are as follows:
                grad (np.ndarray): vector gradients of the POSITIVE log-pdf with respect to param. shape=(d, 1), where
                d is the dimensionality of param.
                hessian (np.ndarray): array of NEGATIVE second derivatives of the log-pdf with respect to param.
                shape=(d, d), where d is the dimensionality of param.

        """
        grad = self.grad_log_p_diff(state=state, param=param)
        if hessian_required:
            hessian = self.hessian_log_p_diff(state=state, param=param)
            return grad, hessian
        return grad

    def grad_log_p_diff(self, state: dict, param: str, step_size: float = 1e-4) -> np.ndarray:
        """Compute vector of derivatives of the POSITIVE log-pdf (with respect to param) using central differences.

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            step_size (float, optional): step size to use for the finite difference derivatives. Defaults to 1e-4.

        Returns:
            (np.ndarray): vector of log-pdf gradients with respect to param. shape=(d, 1), where d is the dimensionality
                of param.

        """
        n_param = np.prod(state[param].shape)
        grad_param = np.full(shape=n_param, fill_value=np.nan)
        for k in range(n_param):
            state_plus = deepcopy(state)
            state_minus = deepcopy(state)

            if sparse.issparse(state[param]):
                m, n = state[param].shape
                step_temp = sparse.csr_array(
                    (np.array([step_size / 2]), np.unravel_index(np.array([k]), (m, n))), shape=(m, n)
                )
                state_plus[param] = state_plus[param] + step_temp
                state_minus[param] = state_minus[param] - step_temp
            else:
                state_plus[param][np.unravel_index(k, state[param].shape)] += step_size / 2
                state_minus[param][np.unravel_index(k, state[param].shape)] += -step_size / 2

            log_p_plus = self.log_p(state=state_plus)
            log_p_minus = self.log_p(state=state_minus)

            grad_param[k] = (log_p_plus - log_p_minus) / step_size
        return grad_param.reshape(state[param].shape)

    def hessian_log_p_diff(self, state: dict, param: str, step_size: float = 1e-4) -> np.ndarray:
        """Compute Hessian matrix of second derivatives of the NEGATIVE log-pdf (with respect to param) using finite differences.

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            step_size (float, optional): step size to use for the finite difference derivatives. Defaults to 1e-4

        Returns:
            (np.ndarray): matrix of log-pdf second derivatives with respect to param. shape=(d, d), where d is the
                dimensionality of param.

        """
        n_param = np.prod(state[param].shape)
        hess_param = np.full(shape=(n_param, n_param), fill_value=np.nan)
        for k in range(n_param):
            state_plus = deepcopy(state)
            state_minus = deepcopy(state)

            if sparse.issparse(state[param]):
                m, n = state[param].shape
                step_temp = sparse.csr_array(
                    (np.array([step_size / 2]), np.unravel_index(np.array([k]), (m, n))), shape=(m, n)
                )
                state_plus[param] = state_plus[param] + step_temp
                state_minus[param] = state_minus[param] - step_temp
            else:
                state_plus[param][np.unravel_index(k, state[param].shape)] += step_size / 2
                state_minus[param][np.unravel_index(k, state[param].shape)] += -step_size / 2

            grad_p_plus = self.grad_log_p(state_plus, param, hessian_required=False)
            grad_p_minus = self.grad_log_p(state_minus, param, hessian_required=False)

            hess_param[:, k] = (grad_p_minus - grad_p_plus).flatten() / step_size

        return hess_param


@dataclass
class Gamma(Distribution):
    """Gamma distribution class defined using shape and rate convention.

    f(x) = x^(shape-1) * exp(-rate*x) * rate^shape / Gamma(shape)

    Attributes:
        shape (Union[str, Identity, LinearCombination, MixtureParameterVector]): Gamma shape parameter.
        rate (Union[str, Identity, LinearCombination, MixtureParameterVector]): Gamma rate parameter.

    """

    shape: Union[str, Identity, LinearCombination, MixtureParameterVector]
    rate: Union[str, Identity, LinearCombination, MixtureParameterVector]

    def __post_init__(self):
        """Parse any str parameter inputs as Parameter.Identity, and check the parameter types."""
        if isinstance(self.shape, str):
            self.shape = Identity(self.shape)

        if not isinstance(self.shape, (Identity, LinearCombination, MixtureParameterVector)):
            raise TypeError("shape expected to be one of [Identity, LinearCombination, MixtureParameterVector]")

        if isinstance(self.rate, str):
            self.rate = Identity(self.rate)

        if not isinstance(self.rate, (Identity, LinearCombination, MixtureParameterVector)):
            raise TypeError("rate expected to be one of [Identity, LinearCombination, MixtureParameterVector]")

    @property
    def _dist_params(self) -> list:
        """Get list of parameter labels across all Parameter objects in distribution (EXCLUDING the response).

        Returns:
            (list): list of parameter labels.

        """
        lst = self.shape.get_param_list() + self.rate.get_param_list()
        return lst

    def log_p(self, state: dict, by_observation: bool = False) -> Union[np.ndarray, float]:
        """Compute the log of the probability density (for current parameter settings).

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of parameters; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p parameters of
                the distribution separately. Defaults to False.

        Returns:
            (Union[np.ndarray, float]): POSITIVE log-density evaluated using the supplied state dictionary.

        """
        log_p = np.sum(
            stats.gamma.logpdf(state[self.response], self.shape.predictor(state), scale=1 / self.rate.predictor(state)),
            axis=0,
        )
        if not by_observation:
            log_p = np.sum(log_p)
        return log_p

    def rvs(self, state, n: int = 1) -> np.ndarray:
        """Generate random samples from the Gamma distribution.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (np.ndarray): random variables generated from distribution returned as p x n where p is the
                dimensionality of the response.

        """
        shape = self.shape.predictor(state)
        rate = self.rate.predictor(state)
        p = max(shape.shape[0], rate.shape[0])
        return stats.gamma.rvs(shape, scale=1 / rate, size=(p, n))


@dataclass
class Categorical(Distribution):
    """Categorical distribution: equivalent to a single trial of a multinomial distribution.

    A 2-category categorical distribution is equivalent to a Bernoulli distribution.

    The response of this distribution is a category index in {0, 1, 2,..., n_cat}: thus, state[self.response] is
    expected to be a np.array with dtype=int. As per other distributions, the expected shape of state[self.response]
    is (p, n), where p=dimensionality of response, and n=number of replicates.

    The prior probability parameter is expected to be a np.ndarray with shape=(p, n_cat).

    Attributes:
        prob (Identity, str): allocation probability parameter.

    """

    prob: Union[str, Identity]

    def __post_init__(self):
        """Parse any str parameter inputs as Parameter.Identity(), and check the parameter types."""
        if isinstance(self.prob, str):
            self.prob = Identity(self.prob)

        if not isinstance(self.prob, Identity):
            raise TypeError("prob expected to be Identity")

    @property
    def _dist_params(self) -> list:
        """Get list of parameter labels across all Parameter objects in distribution (EXCLUDING the response).

        Returns:
            (list): list of parameter labels.

        """
        return self.prob.get_param_list()

    def log_p(self, state: dict, by_observation: bool = False) -> np.ndarray:
        """Compute the log of the probability density (for current parameter settings).

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of responses; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p responses of
                the distribution separately. Defaults to False.

        Returns:
            (Union[np.ndarray, float]): POSITIVE log-density evaluated using the supplied state dictionary.

        """
        n_categories = self.prob.predictor(state).shape[1]
        n = state[self.response].shape[1]

        if n > 1:
            x = np.atleast_3d(state[self.response])
            x = np.equal(np.transpose(x, (0, 2, 1)), np.atleast_3d(range(n_categories)))
        else:
            x = state[self.response] == range(n_categories)

        if by_observation:
            if n > 1:
                prob = np.transpose(np.atleast_3d(self.prob.predictor(state)), (0, 2, 1))
                log_p = stats.multinomial.logpmf(np.transpose(x, (0, 2, 1)), n=1, p=prob)
            else:
                log_p = stats.multinomial.logpmf(x, n=1, p=self.prob.predictor(state))
        else:
            if n > 1:
                x = np.sum(x, axis=2)
            log_p = stats.multinomial.logpmf(x, n=n, p=self.prob.predictor(state))

        return np.sum(log_p, axis=0)

    def rvs(self, state, n: int = 1) -> np.ndarray:
        """Generate a random sample from the distribution.

        Args:
            state (dict): dictionary object containing the current state information
            n (int, optional): specifies the number of random variables required. Defaults to 1

        Returns:
            (np.ndarray): random sample from the categorical distribution. shape=(p, n)

        """
        prob = self.prob.predictor(state)

        d, _ = prob.shape

        cat = np.empty((d, n))
        for i in range(d):
            Z = stats.multinomial.rvs(n=1, p=prob[i, :], size=n)
            _, cat[i, :] = np.nonzero(Z)

        return cat


@dataclass
class Uniform(Distribution):
    """Uniform distribution class for a p-dimensional hyper-rectangle.

    Attributes:
        domain_response_lower (Union[float, np.ndarray]): shape=(p, 1): lower limits for uniform distribution in each
            dimension. Defaults to 0.0.
        domain_response_upper (Union[float, np.ndarray]): shape=(p, 1) upper limits for uniform distribution in each
            dimension. Defaults to 1.0.

    """

    domain_response_lower: Union[float, np.ndarray] = 0.0
    domain_response_upper: Union[float, np.ndarray] = 1.0

    def __post_init__(self):
        """Convert any domain limits supplied as floats to np.ndarray."""
        self.domain_response_lower = np.array(self.domain_response_lower, ndmin=2)
        if self.domain_response_lower.shape[0] == 1:
            self.domain_response_lower = self.domain_response_lower.T
        self.domain_response_upper = np.array(self.domain_response_upper, ndmin=2)
        if self.domain_response_upper.shape[0] == 1:
            self.domain_response_upper = self.domain_response_upper.T

    @property
    def _dist_params(self) -> list:
        """Uniform distribution doesn't have parameters, so return an empty list."""
        return []

    def domain_range(self, state) -> np.ndarray:
        """Get the domain range (upper-lower) from domain_limits.

        Args:
            state (dict): dictionary with current state information.

        Returns:
            (np.ndarray): domain range. shape=(p, 1).

        """
        d = state[self.response].shape[0]
        domain_range = self.domain_response_upper - self.domain_response_lower
        if domain_range.size == 1:
            domain_range = np.ones((d, 1)) * domain_range
        return domain_range

    def log_p(self, state: dict, by_observation: bool = False) -> Union[np.ndarray, float]:
        """Compute the log of the probability density (for current parameter settings).

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of responses; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p responses of
                the distribution separately. Defaults to False.

        Returns:
            (Union[np.ndarray, float]): POSITIVE log-density evaluated using the supplied state dictionary.

        """
        n = state[self.response].shape[1]
        log_p = -np.sum(np.log(self.domain_range(state)))
        if by_observation:
            log_p = np.ones(n) * log_p
        else:
            log_p = n * log_p
        return log_p

    def rvs(self, state, n: int = 1) -> np.ndarray:
        """Generate random samples from the distribution.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (np.ndarray): random variables generated from distribution returned as p x n where p is the
                dimensionality of the response.

        """
        standard_unif = np.random.rand(state[self.response].shape[0], n)
        return self.domain_response_lower + self.domain_range(state) * standard_unif


@dataclass
class Poisson(Distribution):
    """Poisson distribution for count data.

    Attributes:
        rate (Union[str, Identity, LinearCombination, MixtureParameterVector]): Poisson rate parameter.

    """

    rate: Union[str, Identity, LinearCombination, MixtureParameterVector]

    def __post_init__(self):
        """Parse any str parameter inputs as Parameter.Identity, and check the parameter types."""
        if isinstance(self.rate, str):
            self.rate = Identity(self.rate)

        if not isinstance(self.rate, (Identity, LinearCombination, MixtureParameterVector)):
            raise TypeError("rate expected to be one of [Identity, LinearCombination, MixtureParameterVector]")

    @property
    def _dist_params(self) -> list:
        """Get list of parameter labels across all Parameter objects in distribution (EXCLUDING the response).

        Returns:
            (list): list of parameter labels.

        """
        return self.rate.get_param_list()

    def log_p(self, state: dict, by_observation: bool = False) -> np.ndarray:
        """Compute the log of the probability density (for current parameter settings).

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of parameters; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p parameters of
                the distribution separately. Defaults to False.

        Returns:
            (Union[np.ndarray, float]): POSITIVE log-density evaluated using the supplied state dictionary.

        """
        rate = self.rate.predictor(state)
        logpmf = np.sum(stats.poisson.logpmf(state[self.response], rate), axis=0)
        if not by_observation:
            logpmf = np.sum(logpmf)
        return logpmf

    def rvs(self, state: dict, n: int = 1) -> np.ndarray:
        """Generate random samples from the Poisson distribution.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (np.ndarray): random variables generated from distribution returned as p x n where p is the
                dimensionality of the response.

        """
        rate = self.rate.predictor(state)
        return stats.poisson.rvs(mu=rate, size=(rate.shape[0], n))



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""LocationScale module.

This module provides a class definition of the LocationScale class an abstract base class for distributions defined by a
mean and a precision such as the Normal and Lognormal.

"""

from abc import ABC
from dataclasses import dataclass
from typing import Tuple, Union

import numpy as np
from scipy import sparse

from openmcmc import gmrf
from openmcmc.distribution.distribution import Distribution
from openmcmc.parameter import (
    Identity,
    LinearCombination,
    MixtureParameterMatrix,
    MixtureParameterVector,
    ScaledMatrix,
)


@dataclass
class LocationScale(Distribution, ABC):
    """Abstract base class for distributions defined by a mean and a precision such as the Normal and Lognormal.

    Attributes:
       mean (Union[str, Identity, LinearCombination, MixtureParameterVector]): mean parameter (of class Parameter).
       precision (Union[str, Identity, ScaledMatrix, MixtureParameterMatrix]): precision parameter (of class Parameter).

    """

    mean: Union[str, Identity, LinearCombination, MixtureParameterVector]
    precision: Union[str, Identity, ScaledMatrix, MixtureParameterMatrix]

    @property
    def _dist_params(self) -> list:
        """Return the full list of state elements used in the mean and precision parameters."""
        lst = self.mean.get_param_list() + self.precision.get_param_list()
        return lst

    def __post_init__(self):
        """Parse any str parameter inputs as Parameter classes."""
        if isinstance(self.mean, str):
            self.mean = Identity(self.mean)

        if not isinstance(self.mean, (Identity, LinearCombination, MixtureParameterVector)):
            raise TypeError("mean expected to be one of [Identity, LinearCombination, MixtureParameterVector]")

        if isinstance(self.precision, str):
            self.precision = Identity(self.precision)

        if not isinstance(self.precision, (Identity, ScaledMatrix, MixtureParameterMatrix)):
            raise TypeError("precision expected to be one of [Identity, ScaledMatrix, MixtureParameterMatrix]")


class NullDistribution(LocationScale):
    """Null distribution, which returns 0 for the log-likelihood, a zero vector for the gradient and a zero matrix for the Hessian.

    Used in prior recovery testing for reversible jump sampler.

    """

    def log_p(self, state: dict, by_observation: bool = False) -> float:
        """Null log-density function: returns 0.

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of responses; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p responses of
                the distribution separately. Defaults to False.

        Returns:
            (float): 0.0.

        """
        return 0.0

    def grad_log_p(
        self, state: dict, param: str, hessian_required: bool = True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Null gradient function returning an all-zero vector for the gradient, and an all-zero matrix for the Hessian.

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            hessian_required (bool): flag for whether the Hessian should be calculated and supplied as an output.

        Returns:
            (Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]): if hessian_required=True, then a tuple of (gradient,
                hessian) is returned. If hessian_required=False, then just a gradient vector is returned. The returned
                values are as follows:
                grad (np.ndarray): all-zero vector. shape=(d, 1), where d is the dimensionality of param.
                hessian (np.ndarray): all-zero matrix. shape=(d, d), where d is the dimensionality of param.

        """
        if hessian_required:
            return np.zeros(state[param].shape), np.zeros((state[param].shape[0], state[param].shape[0]))

        return np.zeros(state[param].shape)

    def rvs(self, state: dict, n: int = 1) -> None:
        """Null random sampling function.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (None): simply returns None value.

        """
        return None


@dataclass
class Normal(LocationScale):
    """Multivariate normal distribution class.

    Supports both standard multivariate normal and truncated normal distribution cases. By default, no truncation is
    assumed. To truncate the distribution, one or both of self.domain_response_lower or self.domain_response_upper must
    be specified.

    Attributes:
        domain_response_lower (np.array, optional): check lower bound domain to implement truncated sampling. Defaults
            to None.
        domain_response_upper (np.array, optional): check upper bound domain to implement truncated sampling. Defaults
            to None.

    """

    domain_response_lower: np.ndarray = None
    domain_response_upper: np.ndarray = None

    def log_p(self, state: dict, by_observation: bool = False) -> Union[np.ndarray, float]:
        """Compute the log of the probability density for a given state.

        NOTE: This function simply computes the non-truncated Gaussian density: i.e. the extra normalization for the
        truncation is NOT accounted for. Relative densities (differences of log-probabilities) are still valid when
        comparing different response parameter values (with fixed mean and precision parameter values). Comparisons
        for different mean or precision parameters are not valid, since such changes would affect the normalization.

        Args:
            state (dict): dictionary object containing the current parameter information.
            by_observation (bool, optional): indicates whether log-density should be computed for each individual
                response in the distribution. Defaults to False (i.e. the overall log-density is computed).

        Returns:
            (Union[np.ndarray, float]): log-density computed using the values in state.

        """
        Q = self.precision.predictor(state)
        mu = self.mean.predictor(state)
        if self.check_domain_response(state):
            return -np.inf
        log_p = gmrf.multivariate_normal_pdf(x=state[self.response], mu=mu, Q=Q, by_observation=by_observation)
        return log_p

    def check_domain_response(self, state: dict) -> bool:
        """Checks whether the distributional response lies OUTSIDE the defined limits.

        Returns True if the current value of self.response in the supplied state lies OUTSIDE the stated domain;
        returns False otherwise.

        Args:
            state (dict): dictionary object containing the current parameter information.

        Returns:
            (bool): True when the response lies OUTSIDE the valid response domain; False when it lies INSIDE.

        """
        if self.domain_response_lower is not None:
            if np.any(state[self.response] < self.domain_response_lower):
                return True
        if self.domain_response_upper is not None:
            if np.any(state[self.response] > self.domain_response_upper):
                return True
        return False

    def grad_log_p(
        self, state: dict, param: str, hessian_required: bool = True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Gradient and Hessian of the log-Gaussian density, with respect to a given parameter.

        See also distribution.grad_log_p() for more information.

        Handles three possibilities:
            1) param is the response of the distribution, in which case the standard gradient of the log-density is
                returned.
            2) param is a parameter used in the computation of the mean (through a parameter object) and not in the
                computation of the precision, in which case the gradient is computed through application of the chain
                rule. Note that the Hessian calculated in this case is only valid if the dependence of self.mean on
                param is linear.
            3) neither of the above conditions is True, in which case the default finite-difference gradient is
                calculated (using self.grad_log_p_diff() and self.hessian_log_p_diff()). Note that as per those
                docstrings, it is only possible to compute gradients with respect to scalar or vector parameters.

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            hessian_required (bool): flag for whether the Hessian should be calculated and supplied as an output.

        Returns:
            (Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]): if hessian_required=True, then a tuple of (gradient,
                hessian) is returned. If hessian_required=False, then just a gradient vector is returned. The returned
                values are as follows:
                grad (np.ndarray): vector gradients of the POSITIVE log-pdf with respect to param. shape=(n_param, 1)
                hessian (np.ndarray): array of NEGATIVE second derivatives of the log-pdf with respect to param.
                shape=(n_param, n_param)

        """
        if param in self.response:
            Q = self.precision.predictor(state)
            r = state[self.response] - self.mean.predictor(state)
            grad = -Q @ r
            if hessian_required:
                hessian = Q
                if state[param].shape[1] > 1 and sparse.issparse(Q):
                    hessian = sparse.kron(Q, sparse.eye(state[param].shape[1]))
                elif state[param].shape[1] > 1:
                    hessian = np.kron(Q, np.eye(state[param].shape[1]))
                return grad, hessian

        elif param in self.mean.get_grad_param_list() and param not in self.precision.get_grad_param_list():
            Q = self.precision.predictor(state)
            r = np.sum(state[self.response] - self.mean.predictor(state), axis=1, keepdims=True)
            grad_param = self.mean.grad(state, param)
            grad_times_prec = grad_param @ Q
            grad = grad_times_prec @ r
            if hessian_required:
                hessian = state[self.response].shape[1] * grad_times_prec @ grad_param.T
                return grad, hessian

        else:
            grad = self.grad_log_p_diff(state, param)
            if hessian_required:
                hessian = self.hessian_log_p_diff(state, param)
                return grad, hessian

        return grad

    def rvs(self, state: dict, n: int = 1) -> np.ndarray:
        """Generate random samples from the multivariate Gaussian distribution.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (np.ndarray): random variables generated from distribution returned as p x n where p is the
                dimensionality of the response.

        """
        mean = self.mean.predictor(state)
        precision = self.precision.predictor(state)

        if self.domain_response_lower is None and self.domain_response_upper is None:
            return gmrf.sample_normal(mu=mean, Q=precision, n=n)

        return gmrf.sample_truncated_normal(
            mu=mean, Q=precision, lower=self.domain_response_lower, upper=self.domain_response_upper, n=n
        )


@dataclass
class LogNormal(LocationScale):
    """Multivariate log-normal distribution class."""

    def log_p(self, state: dict, by_observation: bool = False) -> np.ndarray:
        """Compute the log of the probability density (for current parameter settings).

        Args:
            state (dict): dictionary object containing the current state information. state[distribution.response]
                is expected to be p x n where: p is the number of responses; n is the number of independent
                replicates/observations.
            by_observation (bool, optional): If True, the log-likelihood is returned for each of the p responses of
                the distribution separately. Defaults to False.

        Returns:
            (Union[np.ndarray, float]): POSITIVE log-density evaluated using the supplied state dictionary.

        """
        Q = self.precision.predictor(state)
        mu = self.mean.predictor(state)
        log_p = gmrf.multivariate_normal_pdf(x=np.log(state[self.response]), mu=mu, Q=Q, by_observation=True) - np.sum(
            np.log(state[self.response]), axis=0
        )
        if not by_observation:
            log_p = np.sum(log_p)
        return log_p

    def grad_log_p(
        self, state: dict, param: str, hessian_required: bool = True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Generate vector of derivatives of the log-pdf with respect to a given parameter, and if required, also generate the Hessian.

        See also distribution.grad_log_p() for more information.

        Handles 3 possibilities:
            1) param is the response of the distribution, in which case the standard gradient of the log-density is
                returned.
            2) param is a parameter used in the computation of the mean (through a parameter object) and not in the
                computation of the precision, in which case the gradient is computed through application of the chain
                rule. Note that the Hessian calculated in this case is only valid if the dependence of self.mean on
                param is linear.
            3) neither of the above conditions is True, in which case the default finite-difference gradient is
                calculated (using self.grad_log_p_diff() and self.hessian_log_p_diff()). Note that as per those
                docstrings, it is only possible to compute gradients with respect to scalar or vector parameters.

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            hessian_required (bool): flag for whether the Hessian should be calculated and supplied as an output.

        Returns:
            (Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]): if hessian_required=True, then a tuple of (gradient,
                hessian) is returned. If hessian_required=False, then just a gradient vector is returned. The returned
                values are as follows:
                grad (np.ndarray): vector gradients of the POSITIVE log-pdf with respect to param. shape=(d, 1), where
                d is the dimensionality of param.
                hessian (np.ndarray): array of NEGATIVE second derivatives of the log-pdf with respect to param.
                shape=(d, d), where d is the dimensionality of param.

        """
        Q = self.precision.predictor(state)
        if param in self.response:
            r = np.log(state[self.response]) - self.mean.predictor(state)
            grad = -(1 / state[self.response]) * (1 + Q @ r)
        elif param in self.mean.get_grad_param_list() and param not in self.precision.get_grad_param_list():
            r = np.sum(np.log(state[self.response]) - self.mean.predictor(state), axis=1, keepdims=True)
            grad_param = self.mean.grad(state, param)
            grad = grad_param @ Q @ r
        else:
            grad = self.grad_log_p_diff(state, param)

        if hessian_required:
            hessian = self.hessian_log_p(state, param)
            return grad, hessian

        return grad

    def hessian_log_p(self, state: dict, param: str) -> np.ndarray:
        """Compute Hessian of the log-density with respect to a given parameter.

        Handles 3 possibilities:
            1) param is the response of the distribution, in which case the Hessian of the log-density is computed
                directly.
            2) param is a parameter used in the computation of the mean (through a parameter object) and not in the
                computation of the precision, and the dependence of the mean parameter on param is linear. The chain
                rule is used to determine the Hessian.
            3) neither of the above conditions is True, in which case the default finite-difference gradient is
                calculated (using self.hessian_log_p_diff()). Note that as per the docstring of
                self.hessian_log_p_diff(), it is only possible to compute gradients with respect to scalar or vector
                parameters.

        NOTE: sparse implementation of response hessian currently converts Q from sparse.

        Args:
            state (dict): contains current state information.
            param (str): name of the parameter for which we compute derivatives.

        Returns:
            (np.ndarray): Hessian of log-density wrt the specified param.

        """
        if param in self.response:
            Q = self.precision.predictor(state)
            r = np.log(state[self.response]) - self.mean.predictor(state)
            reciprocal = 1 / state[self.response]

            if sparse.issparse(Q):
                hess_p = -sparse.diags((np.power(reciprocal, 2) * (1 + Q @ r)).flatten(), offsets=0)
                Q = Q.toarray()
            else:
                hess_p = -np.diagflat(np.power(reciprocal, 2) * (1 + Q @ (r)))

            dim, n = state[self.response].shape
            out = np.zeros((n, dim, n, dim))
            diag = np.einsum("ijik->ijk", out)
            np.einsum("ik, ij, jk -> kij", reciprocal, Q, reciprocal, out=diag)
            out = out.transpose((1, 0, 3, 2))
            out = out.reshape((n * dim, n * dim))
            hess_p = out + hess_p

        elif param in self.mean.get_grad_param_list() and param not in self.precision.get_grad_param_list():
            Q = self.precision.predictor(state)
            grad_param = self.mean.grad(state, param)
            hess_p = state[self.response].shape[1] * grad_param @ Q @ grad_param.T
        else:
            hess_p = self.hessian_log_p_diff(state, param)

        return hess_p

    def rvs(self, state: dict, n: int = 1) -> np.ndarray:
        """Generate random samples from the multivariate log-Gaussian distribution.

        Args:
            state (dict): dictionary object containing the current state information.
            n (int, optional): specifies the number of replicate samples required. Defaults to 1.

        Returns:
            (np.ndarray): random variables generated from distribution returned as p x n where p is the
                dimensionality of the response.

        """
        mean = self.mean.predictor(state)
        precision = self.precision.predictor(state)
        return np.exp(gmrf.sample_normal(mu=mean, Q=precision, n=n))



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Distribution module."""

__all__ = [
    "distribution",
    "location_scale",
]



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Gmrf: gaussian Markov Random Field.

Reference: Rue, Held 2005 Gaussian Markov Random Fields

Helper functions for sampling and dealing with Multivariate normal distributions
defined by precision matrices which avoid the need for direct inversion and efficiently
reuse cholesky factorisations with sparse implementations

Notation:
b: conditional mean
Q: precision matrix
L: lower triangle cholesky factorisation of a precision matrix Q

"""

from typing import Union

import numpy as np
from pandas.arrays import DatetimeArray
from scipy import linalg, sparse
from scipy.sparse import linalg as sparse_linalg
from scipy.stats import truncnorm


def sample_normal(
    mu: np.ndarray, Q: Union[np.ndarray, sparse.csc_matrix] = None, L: np.ndarray = None, n: int = 1
) -> np.ndarray:
    """Generate multivariate random variables from a precision matrix Q using lower cholesky factorisation to get L.

    Note: sparse_linalg.spsolve_triangular compared to sparse_linalg.spsolve, and
    it appears to be much slower.
    Algorithm 2.4 from Rue, Held 2005 Gaussian Markov Random Fields
     Sampling x ~ N(mu , Q^-1)
    1: Compute the lower Cholesky factorisation, Q = L @ L'
    2: Sample z ~ N(0,  I)
    3: Solve L' v = z
    4: Compute x = z + v
    5: Return x
    Args:
        mu (np.array): p x 1 mean
        Q (np.array, optional): p x p for precision matrix. Defaults to None.
        L (np.array, optional): p x p for lower triangular cholesky factorisation of
                                     precision matrix. Defaults to None.
        n (int, optional): number of samples. Defaults to 1.

    Returns:
        (np.array): p x n random normal values

    """
    size = [np.size(mu), n]

    z = np.random.standard_normal(size=size)

    if L is None:
        L = cholesky(Q)

    return solve(L.T, z).reshape(z.shape) + mu


def sample_truncated_normal(
    mu: np.ndarray,
    Q: Union[np.ndarray, sparse.csc_matrix] = None,
    L: np.ndarray = None,
    lower: np.ndarray = None,
    upper: np.ndarray = None,
    n: np.array = 1,
    method="Gibbs",
) -> np.ndarray:
    """Sample from multivariate truncated normal using either rejection sampling or Gibbs sampling.

    Gibbs sampling should be faster but is generated through a markov chain so samples may not be completely independent
    The Markov chain is set up for sampling from gibbs_canonical_truncated_normal which is thinned by every 10
    observations to get more i.i.d. samples

    Rejection sampling will work well for low dimensions and low amounts of truncated but will scale very poorly.

    Args:
        mu (np.array): p x 1 mean
        Q (np.array, optional): p x p for precision matrix. Defaults to None.
        L (np.array, optional): p x p for lower triangular cholesky factorisation of
                                     precision matrix. Defaults to None.
        lower (np.array, optional): lower bound
        upper (np.array, optional): upper bound
        n (int, optional): number of samples. Defaults to 1.
        method (str, optional): defines method to use for TN sampling Either 'Gibbs' or 'Rejection' Defaults to 'Gibbs'.

    Returns:
        (np.array): p x n random truncated normal values

    """
    if method == "Gibbs":
        d = mu.shape[0]
        b = Q @ mu
        Z = np.empty(shape=(d, n))
        Z[:, 0] = sample_truncated_normal_rejection(mu=mu, Q=Q, L=L, lower=lower, upper=upper, n=1).flatten()
        thin = 10
        for i in range(n - 1):
            x = Z[:, i].reshape(d, 1)
            for _ in range(thin):
                x = gibbs_canonical_truncated_normal(b=b, Q=Q, x=x, lower=lower, upper=upper)
                Z[:, i + 1] = x.flatten()
        return Z
    if method == "Rejection":
        return sample_truncated_normal_rejection(mu=mu, Q=Q, L=L, lower=lower, upper=upper, n=n)

    raise TypeError("method should be either Gibbs or Rejection")


def sample_truncated_normal_rejection(
    mu: np.ndarray,
    Q: Union[np.ndarray, sparse.csc_matrix] = None,
    L: np.array = None,
    lower: np.ndarray = None,
    upper: np.ndarray = None,
    n: np.array = 1,
) -> np.array:
    """Sample from multivariate truncated normal using rejection sampling.

    Rejection sampling will work well for low dimensions and low amounts of truncated but will scale very poorly.

    Args:
        mu (np.array): p x 1 mean
        Q (np.array, optional): p x p for precision matrix. Defaults to None.
        L (np.array, optional): p x p for lower triangular cholesky factorisation of
                                     precision matrix. Defaults to None.
        lower (np.array, optional): lower bound
        upper (np.array, optional): upper bound
        n (int, optional): number of samples. Defaults to 1.

    Returns:
        (np.array): p x n random truncated normal values

    """
    if L is None:
        L = cholesky(Q)

    n_bad = n

    if lower is None:
        lower = -np.inf

    if upper is None:
        upper = np.inf

    if np.any(lower >= upper):
        raise ValueError("Error lower bound must be strictly less than upper bound")

    samples = sample_normal(mu, L=L, n=n_bad)
    ind_bad = np.any(np.bitwise_or(samples < lower, samples > upper), axis=0)
    n_bad = np.sum(ind_bad)

    while n_bad > 0:
        sample_temp = sample_normal(mu, L=L, n=n_bad)

        samples[:, ind_bad] = sample_temp
        ind_bad = np.any(np.bitwise_or(samples < lower, samples > upper), axis=0)

        n_bad = np.sum(ind_bad)

    return samples


def sample_normal_canonical(b: np.ndarray, Q: np.ndarray = None, L: np.ndarray = None) -> np.ndarray:
    """Generate multivariate random variables canonical representation precision matrix using cholesky factorisation.

    Algorithm 2.5 from Rue, Held 2005 Gaussian Markov Random Fields:
    Sampling x ~ N( Q^-1 b, Q^-1)
        1: Compute the Cholesky factorisation, Q = L @ L'
        2: Solve L w = b
        3: Solve L' mu = w
        4: Sample z ~ N(0; I)
        5: Solve L' v = z
        6: Compute x = mu + v
        7: Return x

    Steps 2 and 3 are done in the function cho_solve and the output is thus mu.
    Steps 4, 5 and 6 are the algorithm 2.5 implemented in the function sample_normal

    Args:
        b (np.ndarray): p x 1 conditional mean
        Q (np.ndarray, optional): p x p for precision matrix. Defaults to None.
        L (np.ndarray, optional): p x p for lower triangular cholesky factorisation
                                    of precision matrix. Defaults to None.

    Returns:
        (np.ndarray): p x 1 random normal values

    """
    if L is None:
        L = sparse_cholesky(Q)

    mu = cho_solve((L, True), b).reshape(b.shape)

    return sample_normal(mu, L=L)


def gibbs_canonical_truncated_normal(
    b: np.ndarray,
    Q: Union[np.ndarray, sparse.csc_matrix],
    x: np.ndarray,
    lower: np.ndarray = -np.inf,
    upper: np.ndarray = np.inf,
) -> np.ndarray:
    """Generate truncated multivariate random variables from a precision matrix Q using lower cholesky factorisation to get L based on current state x using Gibbs sampling.

    subject to linear inequality constraints
    lower < X < upper

    Lemma 2.1 from Rue, Held 2005 Gaussian Markov Random Fields
     Sampling x ~ N_c( Q^-1 b , Q^-1)
     x_a | x_b ~ N_c( b_a - Q_ab x_b, Q_aa)

    Args:
        b (np.array): p x 1 mean
        Q (np.array): p x p for precision matrix. Defaults to None.
        x (np.array): p x 1 current state.
        lower (np.array, optional): p x 1 lower bound for each dimension
        upper (np.array, optional): p x 1 upper bound for each dimension

    Returns:
        (np.array): p x 1 random normal values

    """
    if (lower == -np.inf or lower is None) and (upper == np.inf or upper is None):
        return sample_normal_canonical(b, Q)

    if lower is None:
        lower = -np.inf
    if upper is None:
        upper = np.inf

    p = np.size(x)
    temp_limit = np.full(shape=(p, 1), fill_value=np.inf)
    lower = np.maximum(lower, -temp_limit)
    upper = np.minimum(upper, temp_limit)

    if p == 1:
        if sparse.issparse(Q):
            Q = Q.toarray()
        return np.array(truncated_normal_rv(mean=b / Q, scale=1 / np.sqrt(Q), lower=lower, upper=upper), ndmin=2)

    if sparse.issparse(Q):
        Q_diag = Q.diagonal()
    else:
        Q_diag = np.diag(Q)

    for i in range(p):
        Q_ii = Q_diag[i]
        v_i = 1 / Q_ii
        scale_i = np.sqrt(v_i)

        if sparse.issparse(Q):
            cond_mean_i = v_i * (b[i] - Q.getrow(i) @ x + Q_ii * x[i])
        else:
            cond_mean_i = v_i * (b[i] - Q[i, :] @ x + Q_ii * x[i])

        x[i] = truncated_normal_rv(mean=cond_mean_i, scale=scale_i, lower=lower[i], upper=upper[i])

    return x


def truncated_normal_rv(
    mean: np.ndarray, scale: np.ndarray, lower: np.ndarray, upper: np.ndarray, size=1
) -> np.ndarray:
    """Wrapper for scipy.stats.truncnorm.rvs handles cases a, b not standard form.

    Args:
        mean (np.array): p x 1 mean for each dimension
        scale (np.array): p x 1 standard deviation for each dimension
        lower (np.array): p x 1 lower bound for each dimension
        upper (np.array): p x 1 upper bound for each dimension
        size (int): size of output array default = 1

    Returns:
        (np.ndarray): size x 1  truncated normal samples

    """
    if lower is None:
        lower = -np.inf

    if upper is None:
        upper = np.inf

    a, b = (lower - mean) / scale, (upper - mean) / scale
    return truncnorm.rvs(a, b, loc=mean, scale=scale, size=size)


def truncated_normal_log_pdf(
    x: np.ndarray, mean: np.ndarray, scale: np.ndarray, lower: np.ndarray, upper: np.ndarray
) -> np.ndarray:
    """Wrapper for scipy.stats.truncnorm.logpdf handles cases a, b not standard form.

    Args:
        x (np.ndarray): values
        mean (np.ndarray): mean
        scale (np.ndarray): standard deviation
        lower (np.ndarray): lower bound
        upper (np.ndarray): upper bound

    Returns:
        (np.ndarray): truncated normal sample

    """
    if lower is None:
        lower = -np.inf

    if upper is None:
        upper = np.inf

    a, b = (lower - mean) / scale, (upper - mean) / scale
    return truncnorm.logpdf(x, a, b, loc=mean, scale=scale)


def multivariate_normal_pdf(
    x: np.ndarray, mu: np.ndarray, Q: Union[np.ndarray, sparse.csc_matrix], by_observation: bool = False
) -> Union[np.ndarray, float]:
    """Compute diagonalized log-pdf of a multivariate Gaussian distribution in terms of the precision matrix, can take sparse precision matrix inputs.

    Args:
        x (np.ndarray): dim  x n value for the distribution response. where dim is the number of dimensions and n
                        is the number of observations
        mu (np.ndarray): dim x 1 distribution mean vector.
        Q (np.ndarray, sparse.csc_matrix): dim x dim distribution precision matrix can be sparse or np.array
        by_observation (bool, optional): indicates whether we should sum over observations default= False

    Returns:
        (np.ndarray): log-pdf of the Gaussian distribution either:
                                    (1,) if by_observation = False or
                                    (n,) if by_observation = True

    """
    L = cholesky(Q)
    dim = L.shape[0]

    log_det_precision = 2 * np.sum(np.log(L.diagonal()))
    Q_residual = L.T @ (x - mu)
    log_p = (1 / 2) * (log_det_precision - dim * np.log(2 * np.pi) - np.sum(np.power(Q_residual, 2), axis=0))

    if not by_observation:
        log_p = np.sum(log_p)
    return log_p


def precision_temporal(
    time: DatetimeArray, unit_length: float = 1.0, is_sparse: bool = True
) -> Union[np.ndarray, sparse.csc_matrix]:
    """Generate temporal difference penalty matrix.

    Details can be found on pages 97-99 of 'Gaussian Markov Random Fields'
    [Rue, Held 2005], 'The first-order random walk for irregular locations'.

    Converts time to number of seconds then call precision_irregular

    Args:
        time (DatetimeArray): vector of times
        unit_length (float, optional): numbers seconds to define unit difference Defaults to 1 second
        is_sparse (bool, optional): Flag if generated as sparse. Defaults to True.

    Returns:
        P (Union[np.ndarray, sparse.csc_matrix]): un-scaled precision matrix

    """
    s = (time - time.min()).total_seconds() / unit_length

    return precision_irregular(s, is_sparse=is_sparse)


def precision_irregular(s: np.ndarray, is_sparse: bool = True) -> Union[np.ndarray, sparse.csc_matrix]:
    """Generate penalty matrix from irregular observations using first order random walk.

    Details can be found on pages 97-99 of 'Gaussian Markov Random Fields'
    [Rue, Held 2005], 'The first-order random walk for irregular locations'.

    Diagonal and off-diagonal elements of the precision found as follows:
                1/del_{i-1} + 1/del_{i},    j = i,
    Q_{ij} =    -1/del_{i},                 j = i+1,
                0,                          else.
    where del = [t_{i+1} - t_{i}]

    Args:
        s (np.ndarray): vector of locations.
        is_sparse (bool, optional): Flag if generated as sparse. Defaults to True.

    Returns:
        P ( Union[np.ndarray, sparse.csc_matrix]): un-scaled precision matrix

    """
    if s.ndim > 1:
        s = np.squeeze(s)

    if s.size > 1:
        delta_reciprocal = 1.0 / np.diff(s)

        d_0 = np.append(
            np.append(delta_reciprocal[0], delta_reciprocal[:-1] + delta_reciprocal[1:]), delta_reciprocal[-1]
        )
        if is_sparse:
            P = sparse.diags(diagonals=(-delta_reciprocal, d_0, -delta_reciprocal), offsets=[-1, 0, 1], format="csc")
        else:
            P = np.diag(d_0, k=0) - np.diag(delta_reciprocal, k=-1) - np.diag(delta_reciprocal, k=1)
    else:
        P = np.array(1, ndmin=2)

    return P


def solve(
    a: Union[np.ndarray, sparse.csc_matrix], b: Union[np.ndarray, sparse.csc_matrix]
) -> Union[np.ndarray, sparse.csc_matrix]:
    """Solve a linear matrix equation, or system of linear scalar equations.

    Computes the “exact” solution, x, of the well-determined, i.e., full rank, linear matrix equation ax = b.

    If inputs are sparse calls scipy.linalg.spsolve else calls np.linalg.solve

    Args:
        a (Union[np.ndarray, sparse.csc_matrix]): _description_
        b (Union[np.ndarray, sparse.csc_matrix]): _description_

    Returns
        Union(np.ndarray, sparse.csc_matrix) solution to the system in same format as input

    """
    if sparse.issparse(a) or sparse.issparse(b):
        return sparse_linalg.spsolve(a, b)

    return np.linalg.solve(a, b)


def cho_solve(c_and_lower: tuple, b: Union[np.ndarray, sparse.csc_matrix]) -> Union[np.ndarray, sparse.csc_matrix]:
    """Solve the linear equations A x = b, given the Cholesky factorization of A.

    If inputs are sparse calls sparse solvers otherwise uses scipy.linalg.cho_solve

    Args:
        c_and_lower ( tuple(Union(np.ndarray, sparse.csc_matrix), bool)): Cholesky factorization of A
                    and flag for if it is a lower Cholesky
        b (Union(np.ndarray, sparse.csc_matrix)): Right-hand side

    Returns
        (Union(np.ndarray, sparse.csc_matrix)) The solution to the system A x = b

    """
    if sparse.issparse(c_and_lower[0]) or sparse.issparse(b):
        if c_and_lower[1]:
            L = c_and_lower[0]
            U = c_and_lower[0].T
        else:
            L = c_and_lower[0].T
            U = c_and_lower[0]

        w = sparse_linalg.spsolve(L, b)
        return sparse_linalg.spsolve(U, w)

    return linalg.cho_solve(c_and_lower, b)


def cholesky(Q: Union[np.ndarray, sparse.csc_matrix], lower: bool = True) -> Union[np.ndarray, sparse.csc_matrix]:
    """Compute Cholesky factorization of input matrix.

    If it is sparse will use gmf.sparse_cholesky otherwise will use linalg.cholesky

    Args:
        Q (Union[np.ndarray, sparse.csc_matrix]):  precision matrix, for factorization
        lower (bool, optional): flag for lower triangular matrix, default is true

    Returns
       (Union[np.ndarray, sparse.csc_matrix]: Cholesky factorization of the input in the same format as the input

    """
    if sparse.issparse(Q):
        L = sparse_cholesky(Q)
    else:
        L = np.linalg.cholesky(Q)

    if lower:
        return L

    return L.T


def sparse_cholesky(Q: sparse.csc_matrix) -> sparse.csc_matrix:
    """Compute sparse Cholesky factorization of input matrix.

    Uses the scipy.sparse functionality for LU decomposition, and converts
    to Cholesky factorization. Approach taken from:
    https://gist.github.com/omitakahiro/c49e5168d04438c5b20c921b928f1f5d

    If the sparse matrix is identified as unsuitable for Cholesky factorization,
    the function attempts to compute the Chol of the dense matrix instead.

    Args:
        Q (sparse.csc_matrix): sparse precision matrix, for factorization

    Returns:
        (sparse.csc_matrix): Cholesky factorization of the input

    """
    m = Q.shape[0]
    n = Q.shape[1]
    if m != n:
        raise ValueError("Matrix is not square")

    if sparse.issparse(Q):
        if not isinstance(Q, sparse.csc_matrix):
            Q = Q.tocsc()
        fact_lu = sparse_linalg.splu(Q, diag_pivot_thresh=0, options={"RowPerm": False, "ColPerm": False})
        if (fact_lu.U.diagonal() > 0).all():
            return fact_lu.L.dot(sparse.diags(fact_lu.U.diagonal() ** 0.5))

        return np.linalg.cholesky(Q.toarray())

    return np.linalg.cholesky(Q)



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""Model module.

This module provides a class definition of the Model class, a dictionary-like collection of distributions to form a
model.

"""

from dataclasses import dataclass
from typing import Tuple, Union

import numpy as np

from openmcmc.distribution.distribution import Distribution


@dataclass
class Model(dict):
    """Dictionary-like collection of distributions to form a model.

    self.keys() indexes the responses of the distributions in the collection; self.values() contain the individual
    distribution objects in the model, of type Distribution.

    Attributes:
        response (dict): dictionary with keys corresponding to the data values within state, and values corresponding
            to the desired predictor values within the data distributions (for storing fitted values).

    """

    def __init__(self, distributions: list[Distribution], response: dict = None):
        dist_dict = {}
        for dist in distributions:
            dist_dict[dist.response] = dist
        super().__init__(dist_dict)
        self.response = response

    def conditional(self, param: str):
        """Return sub-model which consists of the subset of distributions dependent on the supplied parameter.

        Args:
            param (str): parameter to find within the model distributions.

        Returns:
            (Model): model object containing only distributions which have a dependence on param.

        """
        conditional_dist = []
        for dst in self.values():
            if param in dst.param_list:
                conditional_dist.append(dst)
        return Model(conditional_dist)

    def log_p(self, state: dict) -> Union[float, np.ndarray]:
        """Compute the log-probability density for the full model.

        Args:
            state (dict): dictionary with current state information.

        Returns:
            (Union[float, np.ndarray]): POSITIVE log-probability density evaluated using the information in state.

        """
        log_prob = 0
        for dst in self.values():
            log_prob += dst.log_p(state)
        return log_prob

    def grad_log_p(
        self, state: dict, param: str, hessian_required: bool = True
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Generate vector of derivatives of the log-pdf with respect to a given parameter, as the sum of the derivatives of all the individual components of the model. If required, also generate the Hessian.

        Function only defined for scalar- and vector-valued parameters param. If hessian_required=True, this function
        returns a tuple of (gradient, Hessian). If hessian_required=False, this function returns a np.ndarray (just
        the gradient of the log-density).

        Args:
            state (dict): current state information.
            param (str): name of the parameter for which we compute derivatives.
            hessian_required (bool): flag for whether the Hessian should be calculated and supplied as an output.

        Returns:
            (Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]): if hessian_required=True, then a tuple of (gradient,
                hessian) is returned. If hessian_required=False, then just a gradient vector is returned. The returned
                values are as follows:
                grad (np.ndarray): vector gradients of the POSITIVE log-pdf with respect to param. shape=(d, 1), where
                d is the dimensionality of param.
                hessian (np.ndarray): array of NEGATIVE second derivatives of the log-pdf with respect to param.
                shape=(d, d), where d is the dimensionality of param.

        """
        grad_sum = np.zeros(shape=state[param].shape)
        if hessian_required:
            hessian_sum = np.zeros(shape=(state[param].shape[0], state[param].shape[0]))

        for dist in self.values():
            grad_out = dist.grad_log_p(state, param, hessian_required=hessian_required)
            if hessian_required:
                grad_sum += grad_out[0]
                hessian_sum += grad_out[1]
            else:
                grad_sum += grad_out

        if hessian_required:
            return grad_sum, hessian_sum

        return grad_sum



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""MetropolisHastings module.

This module provides a class definition of the MetropolisHastings class an abstract base class for implementation of
Metropolis-Hastings-type sampling algorithms for a model.

"""

from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Callable, Tuple

import numpy as np
from scipy.stats import norm

from openmcmc import gmrf
from openmcmc.sampler.sampler import MCMCSampler


@dataclass
class AcceptRate:
    """Class for dealing with calculation of acceptance rates.

    Called from MetropolisHastings-type samplers.

    Attributes:
        count: counters of current number of proposals and accepted proposals from a MH chain

    """

    def __init__(self):
        self.count = {"accept": 0, "proposal": 0}

    @property
    def acceptance_rate(self) -> float:
        """Acceptance rate property, as calculated from counters.

        Returns:
            (float): percentage proposals accepted in chain

        """
        return self.count["accept"] / self.count["proposal"] * 100

    def get_acceptance_rate(self) -> str:
        """Return acceptance rate formatted as string.

        Returns:
            (str): acceptance rate string print out

        """
        if self.count["proposal"] == 0:
            return "No proposals"
        return f"Acceptance rate {self.acceptance_rate:.0f}%"

    def increment_accept(self):
        """Increment acceptance count."""
        self.count["accept"] += 1

    def increment_proposal(self):
        """Increment proposal count."""
        self.count["proposal"] += 1


@dataclass
class MetropolisHastings(MCMCSampler):
    """Abstract base class for implementation of Metropolis-Hastings-type sampling algorithms for a model.

    Subclasses include RandomWalk and ManifoldMALA.

    https://en.wikipedia.org/wiki/Metropolis%E2%80%93Hastings_algorithm

    Attributes:
        step (np.ndarray): step size for Metropolis-Hastings proposals. Should either have shape=(p, 1) or shape=(p, n),
            where p is the dimension of the parameter, and n is the number of replicates.
        accept_rate (AcceptRate): Acceptance Rate counter to keep track of proposals.

    """

    step: np.ndarray = field(default_factory=lambda: np.array([0.2], ndmin=2), init=True)
    accept_rate: AcceptRate = field(default_factory=lambda: AcceptRate(), init=False)

    @abstractmethod
    def proposal(self, current_state: dict, param_index: int = None) -> Tuple[dict, float, float]:
        """Method which generates proposed state from current state, and computes corresponding transition probabilities.

        Args:
            current_state (dict): current state
            param_index (int): subset of parameter used in proposal, If none all parameters are used

        Returns:
            prop_state (dict): updated proposal_state dictionary.
            logp_pr_g_cr (float): log-density of proposed state given current state.
            logp_cr_g_pr (float): log-density of current state given proposed state.

        """

    def sample(self, current_state: dict) -> dict:
        """Generate a sample from the specified Metropolis-Hastings-type method.

        https://en.wikipedia.org/wiki/Metropolis%E2%80%93Hastings_algorithm

        generate proposal state x' from  current_state x and accept or reject proposal according to the probability:
        A(x',x) = min(1, (P(x')g(x|x'))/(P(x)g(x'|x)))
        where:
            - P(x) is the probability of the state x
            - g(x|x') is the probability of moving from state x to x'

        The exact method for the proposal (and therefore the form of the proposal distribution) is determined by the
        specific type of MetropolisHastings Sampler used.

        Args:
            current_state (dict): dictionary containing the current sampler state.

        Returns:
            current_state (dict): with updated sample for self.param.

        """
        prop_state, logp_pr_g_cr, logp_cr_g_pr = self.proposal(current_state)
        current_state = self._accept_reject_proposal(current_state, prop_state, logp_pr_g_cr, logp_cr_g_pr)
        return current_state

    def _accept_reject_proposal(
        self, current_state: dict, prop_state: dict, logp_pr_g_cr: float, logp_cr_g_pr: float
    ) -> dict:
        """Accept or Reject Metropolis-Hastings-type proposal.

        Computes the log posterior for the current and proposed states, and evaluates the log acceptance probability.
        Accepts the proposal with probability A(x, x'), and returns either the proposed or the current state
        accordingly.

        Increments self.acceptance_rate() to indicate that a proposal has been made, and also increments the acceptance
        counter if the proposal is subsequently accepted.

        Args:
            current_state (dict): current state dictionary
            prop_state (dict): proposal_state dictionary
            logp_pr_g_cr (float): log posterior of proposal given current state
            logp_cr_g_pr (float): log posterior of current state given proposals

        Returns:
            (dict): updated current state dictionary, after the proposal has either been accepted or rejected.

        """
        self.accept_rate.increment_proposal()
        logp_cs = 0
        logp_pr = 0
        for model in self.model.values():
            logp_cs += model.log_p(current_state)
            logp_pr += model.log_p(prop_state)
        log_accept = logp_pr + logp_cr_g_pr - (logp_cs + logp_pr_g_cr)

        if self.accept_proposal(log_accept):
            current_state = prop_state
            self.accept_rate.increment_accept()
        return current_state

    @staticmethod
    def accept_proposal(log_accept: float) -> bool:
        """Decide to accept or reject proposal based on log acceptance probability.

        Args:
            log_accept (np.float64): log acceptance probability.

        Returns:
            (bool): True for accept, False for Reject.

        """
        return np.log(np.random.rand()) < log_accept


@dataclass
class RandomWalk(MetropolisHastings):
    """Subtype of MetropolisHastings sampler that uses Gaussian random Walk proposals.

    Supports both non-truncated and truncated Gaussian proposals: specifying self.domain limits leads to a truncated
    proposal mechanism.

    Allows for the possibility that other elements of the model state have a dependence on the value of self.param, and
    if so should change when this value changes. If supplied, the self.state_update_function() property is called by the
    proposal function to update any other elements of the state as required.

    Attributes:
        domain_limits (np.ndarray): array with shape=(p, 2), where p is the dimensionality of the parameter being
            sampled. The first column gives the lower limits for the proposal, the second column gives the upper limits.
        state_update_function (Callable): function which updates other elements of proposed state based on the proposed
            value for param.

    """

    domain_limits: np.ndarray = None
    state_update_function: Callable = None

    def __post_init__(self):
        """Derive conditional model instead of storing all distributions where things are simple.

        However, this should not be done in the case where a state_update_function is provided as we don't know  in
        general what/how parameters might change so need to keep full model to avoid incorrect conditioning.

        """
        if self.state_update_function is None:
            self.model = self.model.conditional(self.param)
        self.step = np.array(self.step, ndmin=2)

    def proposal(self, current_state: dict, param_index: int = None) -> Tuple[dict, float, float]:
        """Updates the current value of self.param using a (truncated) Gaussian random walk proposal.

        In the non-truncated case, the proposal mechanism is symmetric, i.e. logp_pr_g_cr = logp_cr_g_pr. In this
        instance, the function simply returns logp_pr_g_cr = logp_cr_g_pr = 0, since these terms would anyway cancel
        in the calculation of the acceptance ratio.

        Introducing a truncation into the proposal distribution means that the proposal is no longer symmetric, and so
        the log-proposal densities are computed in these cases.

        Enables 3 different possibilities for the step size:
            1) shape=(1, 1): scalar step size, identical for every element of the parameter.
            2) shape=(p, 1): step size with the same shape as the parameter being sampled (for one or many replicates).
            3) shape=(p, n): a p-dimensional step size for each of n-replicates.

        Args:
            current_state (dict): dictionary containing current parameter values.
            param_index (int): subset of parameter used in proposal, If none all parameters are used

        Returns:
            prop_state (dict): updated proposal_state dictionary.
            logp_pr_g_cr (float): log-density of proposed state given current state.
            logp_cr_g_pr (float): log-density of current state given proposed state.

        """
        prop_state = deepcopy(current_state)

        if param_index is None:
            mu = prop_state[self.param]
            step = self.step
        else:
            mu = prop_state[self.param][:, param_index]
            if self.step.shape[1] == 1:
                step = self.step.flatten()
            else:
                step = self.step[:, param_index].flatten()

        if self.domain_limits is None:
            z = mu + norm.rvs(size=prop_state[self.param].shape, scale=step)
            logp_pr_g_cr = logp_cr_g_pr = 0.0
        else:
            lb = self.domain_limits[:, 0]
            ub = self.domain_limits[:, 1]
            z = gmrf.truncated_normal_rv(mean=mu, scale=step, lower=lb, upper=ub, size=len(lb))
            logp_pr_g_cr = np.sum(gmrf.truncated_normal_log_pdf(z, mu, step, lower=lb, upper=ub))
            logp_cr_g_pr = np.sum(gmrf.truncated_normal_log_pdf(mu, z, step, lower=lb, upper=ub))

        if param_index is None:
            prop_state[self.param] = z
        else:
            prop_state[self.param][:, param_index] = z

        if callable(self.state_update_function):
            prop_state = self.state_update_function(prop_state, param_index)

        return prop_state, logp_pr_g_cr, logp_cr_g_pr


@dataclass
class RandomWalkLoop(RandomWalk):
    """Subtype of MetropolisHastings sampler which updates each of n replicates of a parameter one-at-a-time, rather than all simultaneously."""

    def sample(self, current_state: dict) -> dict:
        """Update each of n replicates of a given parameter in a loop, rather than simultaneously.

        Args:
            current_state (dict): dictionary containing the current sampler state.

        Returns:
            current_state (dict): with updated sample for self.param.

        """
        for param_index in range(current_state[self.param].shape[1]):
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.proposal(current_state, param_index)
            current_state = self._accept_reject_proposal(current_state, prop_state, logp_pr_g_cr, logp_cr_g_pr)
        return current_state


@dataclass
class ManifoldMALA(MetropolisHastings):
    """Class implementing manifold Metropolis-adjusted Langevin algorithm (mMALA) proposal mechanism.

    Reference: Riemann manifold Langevin and Hamiltonian Monte Carlo methods, Mark Girolami, Ben Calderhead,
    03 March 2011 https://doi.org/10.1111/j.1467-9868.2010.00765.x

    """

    def proposal(self, current_state: dict, param_index: int = None) -> Tuple[dict, np.ndarray, np.ndarray]:
        """Generate mMALA proposed state from current state using gradient and hessian, and compute corresponding log-transition probabilities.

        Args:
            current_state (dict): dictionary containing current parameter values.
            param_index (int): required input from superclass. Not used; defaults to None.

        Returns:
            prop_state (dict): updated proposal_state dictionary.
            logp_pr_g_cr (np.ndarray): log-density of proposed state given current state.
            logp_cr_g_pr (np.ndarray): log-density of current state given proposed state.

        """
        prop_state = deepcopy(current_state)

        mu_cr, chol_cr = self._proposal_params(current_state)
        prop_state[self.param] = gmrf.sample_normal(mu_cr, L=chol_cr)
        logp_pr_g_cr = self._log_proposal_density(prop_state, mu_cr, chol_cr)

        mu_pr, chol_pr = self._proposal_params(prop_state)
        logp_cr_g_pr = self._log_proposal_density(current_state, mu_pr, chol_pr)

        return prop_state, logp_pr_g_cr, logp_cr_g_pr

    def _proposal_params(self, current_state: dict) -> Tuple[np.ndarray, np.ndarray]:
        """Returns the mean vector and the Cholesky factorization of the precision matrix for the mMALA proposal.

        The density for either the forward or return proposal in an mMALA scheme is a Gaussian. In the case of the
        forward proposal, the density is as follows:
            q(prop | theta_0) ~ N(mu*, stp^2 * H ^-1 )
        where:
            mu* = theta_0 + 1/2 * stp^2 * H ^-1 @ G
            H = hessian(theta_0)
            G = gradient(theta_0)

        Args:
            current_state (dict): dictionary containing current parameter values.

        Returns:
            mu_cr (np.ndarray): mean for proposal distribution, shape=(p, 1).
            chol_cr (np.ndarray): lower triangular Cholesky factorization of precision matrix, shape=(p, p).

        """
        grad_cr, hessian_cr = self.model.grad_log_p(current_state, param=self.param, hessian_required=True)
        precision_cr = hessian_cr / (self.step**2)
        chol_cr = gmrf.cholesky(precision_cr)
        mu_cr = current_state[self.param] + (1 / 2) * gmrf.cho_solve((chol_cr, True), grad_cr).reshape(grad_cr.shape)
        return mu_cr, chol_cr

    def _log_proposal_density(self, state: dict, mu: np.ndarray, chol: np.ndarray) -> np.ndarray:
        """Evaluate the log-proposal density for the mMALA transition.

        log determinant calculated using:
        https://blogs.sas.com/content/iml/2012/10/31/compute-the-log-determinant-of-a-matrix.html


        A quadratic form can be expressed in terms of the Cholesky factorization of the matrix as:
            r' Q r = r' L L' r = w' w =sum(w^2)
        where:
            w = L' r
            r = prm - mu

        Args:
            state (dict): dictionary containing parameter values.
            mu (np.ndarray): mean vector, shape=(p, 1).
            chol (np.ndarray): LOWER triangular cholesky factorization of the precision matrix, shape=(p, p)

        Returns:
            (np.ndarray): log-transition probability.

        """
        w = chol.transpose() @ (state[self.param] - mu)
        return np.sum(np.log(chol.diagonal())) - 0.5 * w.T.dot(w)



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
"""ReversibleJump module.

This module provides a class definition of the ReversibleJump class a class for reversible jump sampling for given
parameter and associated parameters.

"""

from copy import deepcopy
from dataclasses import dataclass
from typing import Callable, Tuple, Union

import numpy as np
from scipy.stats import randint, uniform

from openmcmc import gmrf
from openmcmc.sampler.metropolis_hastings import MetropolisHastings


@dataclass
class ReversibleJump(MetropolisHastings):
    """Reversible jump sampling for given parameter and associated parameter.

    self.param corresponds to a number of elements, which will either increase of decrease by 1. self.associated_params
    corresponds to an associated set of self.param parameters, to which we either add or remove an element for a birth
    or death move.

    The attributes self.state_birth_function and self.state_death_function can be used to supply functions which
    implement problem-specific alterations to elements of the state on the occurrence of a birth or death move
    respectively. For example, it may be required to update a basis matrix in the state after a change in the number
    of knots/locations associated with the basis definition.

    The functions self.matched_birth_transition and self.matched_death_transition implement optional functionality which
    can be used to ensure consistency between sets of basis parameters before and after a transition. These work by
    ensuring that the basis predictions before and after the transition match, then applies Gaussian random noise (with
    a given standard deviation) to the coefficient of the new element.

    Attributes:
        associated_params (list or string): a list or a string associated with the dimension jump. List of additional
            parameters that need to be created/removed as part of the dimension change. The default behaviour is to
            sample the necessary additional values from the associated parameter prior distribution. Defaults to None.
        n_max (int): upper limit on self.param (lower limit is assumed to be 1).
        birth_probability (float): probability that a birth move is chosen on any given iteration of the algorithm
            (death_probability = 1 - birth_probability). Defaults to 0.5.
        state_birth_function (Callable): function which implements problem-specific requirements for updates to elements
            of the state as part of a birth function (e.g. updates to a problem-specific basis matrix based given
            additional location parameters). Defaults to None.
        state_death_function (Callable): function which implements problem-specific requirements for updates to elements
            of state as part of a death function. Should mirror the supplied state_birth_function. Defaults to None.
        matching_params (dict): dictionary of parameters required for the matched coefficient transitions- for details
            of what it should contain, see self.matched_birth_transition.

    """

    associated_params: Union[list, str, None] = None
    n_max: Union[int, None] = None
    birth_probability: float = 0.5
    state_birth_function: Union[Callable, None] = None
    state_death_function: Union[Callable, None] = None
    matching_params: Union[dict, None] = None

    def __post_init__(self):
        """Empty function to prevent super.__post_init__ from being run.

        The whole model should be attached in this instance, rather than simply those elements with a dependence on
        self.param.

        """
        if isinstance(self.associated_params, str):
            self.associated_params = [self.associated_params]

    def proposal(self, current_state: dict, param_index: int = None) -> Tuple[dict, float, float]:
        """Make a proposal, and compute related transition probabilities for the move.

        Args:
            current_state (dict): dictionary with current parameter values.
            param_index (int): not used, included for compatibility with superclass.

        Returns:
            prop_state (dict): dictionary updated with proposed value for self.param.
            logp_pr_g_cr (float): transition probability for proposed state given current state.
            logp_cr_g_pr (float): transition probability for current state given proposed state.

        """
        birth = self.get_move_type(current_state)
        if birth:
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.birth_proposal(current_state=current_state)
        else:
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.death_proposal(current_state=current_state)
        return prop_state, logp_pr_g_cr, logp_cr_g_pr

    def birth_proposal(self, current_state: dict) -> Tuple[dict, float, float]:
        """Make a birth proposal move: INCREASES state[self.param] by 1.

        Also makes a proposal for a new element of an associated parameter, state[self.associated_params], by generating a draw
        from the prior distribution for self.associated_params.

        self.state_birth_function() is a function which can be optionally specified for altering the dimensionality of
        any other parameters associated with the dimension change (e.g. a basis matrix, or an allocation parameter).

        If the self.matching_params dictionary is specified, self.matched_birth_transition() is used to generate a
        proposal for a set of basis parameters such that the predicted values match before and after the transition.

        NOTE: log-probability for deletion of a particular knot (-log(n + 1)) is cancelled by the contribution from
        the order statistics densities, log((n + 1)! / n!) = log(n + 1). Therefore, both contributions are omitted from
        the calculation. For further information, see Richardson & Green 1997, Section 3.2:
        https://people.maths.bris.ac.uk/~mapjg/papers/RichardsonGreenRSSB.pdf

        NOTE: log-probability density for the full model is obtained from summing the contribution of the log-density
        for the individual distributions corresponding to each jump parameter.

        Args:
            current_state (dict): dictionary with current parameter values.

        Returns:
            prop_state (dict): dictionary updated with proposed state.
            logp_pr_g_cr (float): transition probability for proposed state given current state.
            logp_cr_g_pr (float): transition probability for current state given proposed state.

        """
        prop_state = deepcopy(current_state)
        prop_state[self.param] = prop_state[self.param] + 1
        log_prop_density = 0

        for associated_key in self.associated_params:
            new_element = self.model[associated_key].rvs(state=current_state, n=1)
            prop_state[associated_key] = np.concatenate((prop_state[associated_key], new_element), axis=1)
            log_prop_density += self.model[associated_key].log_p(current_state, by_observation=True)
        if callable(self.state_birth_function):
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.state_birth_function(current_state, prop_state)
        else:
            logp_pr_g_cr, logp_cr_g_pr = 0.0, 0.0
        if self.matching_params is not None:
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.matched_birth_transition(
                current_state, prop_state, logp_pr_g_cr, logp_cr_g_pr
            )

        p_birth, p_death = self.get_move_probabilities(current_state, True)
        logp_pr_g_cr += np.log(p_birth) + log_prop_density[-1]
        logp_cr_g_pr += np.log(p_death)

        return prop_state, logp_pr_g_cr, logp_cr_g_pr

    def death_proposal(self, current_state: dict) -> Tuple[dict, float, float]:
        """Make a death proposal move: DECREASES state[self.param] by 1.

        Also adjusts the associated parameter state[self.associated_params] by deleting a randomly-selected element.

        self.state_death_function() and self.matched_death_transition() can be used (optional) to specify transitions
        opposite to those used in the birth move.

        NOTE: log-probability density for the full model is obtained from summing the contribution of the log-density
        for the individual distributions corresponding to each jump parameter.

        For further information about the transition, see also self.birth_proposal().

        Args:
            current_state (dict): dictionary with current parameter values.

        Returns:
            prop_state (dict): dictionary updated with proposed state.
            logp_pr_g_cr (float): transition probability for proposed state given current state.
            logp_cr_g_pr (float): transition probability for current state given proposed state.

        """
        prop_state = deepcopy(current_state)
        prop_state[self.param] = prop_state[self.param] - 1
        log_prop_density = 0
        deletion_index = randint.rvs(low=0, high=current_state[self.param])
        for associated_key in self.associated_params:
            prop_state[associated_key] = np.delete(prop_state[associated_key], obj=deletion_index, axis=1)
            log_prop_density += self.model[associated_key].log_p(current_state, by_observation=True)

        if callable(self.state_death_function):
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.state_death_function(
                current_state, prop_state, deletion_index
            )
        else:
            logp_pr_g_cr, logp_cr_g_pr = 0.0, 0.0
        if self.matching_params is not None:
            prop_state, logp_pr_g_cr, logp_cr_g_pr = self.matched_death_transition(
                current_state, prop_state, logp_pr_g_cr, logp_cr_g_pr, deletion_index
            )

        p_birth, p_death = self.get_move_probabilities(current_state, False)
        logp_pr_g_cr += np.log(p_death)
        logp_cr_g_pr += np.log(p_birth) + log_prop_density[-1]

        return prop_state, logp_pr_g_cr, logp_cr_g_pr

    def matched_birth_transition(
        self, current_state: dict, prop_state: dict, logp_pr_g_cr: float, logp_cr_g_pr: float
    ) -> Tuple[dict, float, float]:
        """Generate a proposal for coefficients associated with a birth move, using the principle of matching the predictions before and after the move.

        The parameter vector in the proposed state is computed as: beta* = F @ beta_aug, where:
        F = [G, 0
             0', 1]
        G = (X*' @ X*)^{-1} @ (X*' @ X)
        where X is the original basis matrix, and X* is the augmented basis matrix. For a detailed explanation of the
        approach, see: https://ygraigarw.github.io/ZnnEA1D19.pdf

        The basis matrix in the proposed state should already have been updated in self.state_birth_function(), before
        the call to this function (along with any other associated parameters that need to change shape).

        The following fields should be supplied as part of the self.matching_params dictionary:
            - "variable" (str): reference to the coefficient parameter vector in the state.
            - "matrix" (str): reference to the associated basis matrix in state.
            - "scale" (float): scale of Gaussian noise added to proposal.
            - "limits" (list): [lower, upper] limit for truncated Normal proposals.

        The proposal for the additional basis parameter can be either from:
            - a standard normal distribution (when self.matching_params["limits"] is passed as None).
            - a truncated normal distribution (when self.matching_params["limits"] is a two-element list of the lower
                and upper limits).

        Args:
            current_state (dict): current parameter state as dictionary.
            prop_state (dict): proposed state dictionary, with updated basis matrix.
            logp_pr_g_cr (float): transition probability for proposed state given current state.
            logp_cr_g_pr (float): transition probability for current state given proposed state.

        Returns:
            prop_state (dict): proposed state with updated parameter vector.
            logp_pr_g_cr (float): updated transition probability.
            logp_cr_g_pr (float): updated transition probability.

        """
        vector = self.matching_params["variable"]
        matrix = self.matching_params["matrix"]
        proposal_scale = self.matching_params["scale"]
        proposal_limits = self.matching_params["limits"]

        current_basis = current_state[matrix]
        prop_basis = prop_state[matrix]
        G = np.linalg.solve(
            prop_basis.T @ prop_basis + 1e-10 * np.eye(prop_basis.shape[1]), prop_basis.T @ current_basis
        )
        F = np.concatenate((G, np.eye(N=G.shape[0], M=1, k=-G.shape[0] + 1)), axis=1)
        mu_star = G @ current_state[vector]
        prop_state[vector] = deepcopy(mu_star)

        if proposal_limits is not None:
            prop_state[vector][-1] = gmrf.truncated_normal_rv(
                mean=mu_star[-1], scale=proposal_scale, lower=proposal_limits[0], upper=proposal_limits[1], size=1
            )
            logp_pr_g_cr += gmrf.truncated_normal_log_pdf(
                prop_state[vector][-1], mu_star[-1], proposal_scale, lower=proposal_limits[0], upper=proposal_limits[1]
            )
        else:
            Q = np.array(1 / (proposal_scale**2), ndmin=2)
            prop_state[vector][-1] = gmrf.sample_normal(mu=mu_star[-1], Q=Q, n=1)
            logp_pr_g_cr += gmrf.multivariate_normal_pdf(x=prop_state[vector][-1], mu=mu_star[-1], Q=Q)

        logp_cr_g_pr += np.log(np.linalg.det(F))

        return prop_state, logp_pr_g_cr, logp_cr_g_pr

    def matched_death_transition(
        self, current_state: dict, prop_state: dict, logp_pr_g_cr: float, logp_cr_g_pr: float, deletion_index: int
    ) -> Tuple[dict, float, float]:
        """Generate a proposal for coefficients associated with a death move, as the reverse of the birth proposal in self.matched_birth_transition().

        See self.matched_birth_transition() for further details.

        Args:
            current_state (dict): current parameter state as dictionary.
            prop_state (dict): proposed state dictionary, with updated basis matrix.
            logp_pr_g_cr (float): transition probability for proposed state given current state.
            logp_cr_g_pr (float): transition probability for current state given proposed state.
            deletion_index (int): index of the basis element to be deleted

        Returns:
            prop_state (dict): proposed state with updated parameter vector.
            logp_pr_g_cr (float): updated transition probability.
            logp_cr_g_pr (float): updated transition probability.

        """
        vector = self.matching_params["variable"]
        matrix = self.matching_params["matrix"]
        proposal_scale = self.matching_params["scale"]
        proposal_limits = self.matching_params["limits"]

        current_basis = current_state[matrix]
        prop_basis = prop_state[matrix]
        G = np.linalg.solve(
            current_basis.T @ current_basis + 1e-10 * np.eye(current_basis.shape[1]), current_basis.T @ prop_basis
        )
        F = np.insert(G, obj=deletion_index, values=np.eye(N=G.shape[0], M=1, k=-deletion_index).flatten(), axis=1)
        mu_aug = np.linalg.solve(F, current_state[vector])
        param_del = mu_aug[deletion_index]
        prop_state[vector] = np.delete(mu_aug, obj=deletion_index, axis=0)

        logp_pr_g_cr += np.log(np.linalg.det(F))
        if proposal_limits is not None:
            logp_cr_g_pr += gmrf.truncated_normal_log_pdf(
                param_del, np.array(0), proposal_scale, lower=proposal_limits[0], upper=proposal_limits[1]
            )
        else:
            logp_cr_g_pr += gmrf.multivariate_normal_pdf(
                x=param_del, mu=np.array(0.0, ndmin=2), Q=np.array(1 / (proposal_scale**2), ndmin=2)
            )

        return prop_state, logp_pr_g_cr, logp_cr_g_pr

    def get_move_type(self, current_state: dict) -> bool:
        """Select the type of move (birth or death) to be made at the current iteration.

        Logic for the choice of move is as follows:
            - if state[self.param]=self.n_max, it is not possible to increase self.param, so a death move is chosen.
            - if state[self.param]=1, it is not possible to decrease self.param, so a birth move is chosen.
            - in any other state, a birth move is chosen with probability self.birth_probability, or a death move is
                chosen with probability (1 - self.birth_probability).

        Args:
            current_state (dict): dictionary with current parameter values.

        Returns:
            (bool): if True, make a birth proposal; if False, make a death proposal.

        """
        if current_state[self.param] == self.n_max:
            return False
        if current_state[self.param] == 1:
            return True

        return uniform.rvs() <= self.birth_probability

    def get_move_probabilities(self, current_state: dict, birth: bool) -> Tuple[float, float]:
        """Get the state-dependent probabilities of the forward and reverse moves, accounting for edge cases.

        Returns a tuple of (p_birth, p_death), where these should be interpreted as follows:
            Birth move: p_birth = probability of birth from CURRENT state.
                        p_death = probability of death from PROPOSED state.
            Death move: p_death = probability of death in CURRENT state.
                        p_birth = probability of birth in PROPOSED state.

        In standard cases (away from the limits, assumed to be at [1, n_max]):
            p_birth = q; p_death = 1 - q

        In edge cases (either where we are at one of the limits, or where our chosen move takes us into a limiting
        case), we adjust the probability of either the forward or the reverse move to account for this. E.g.: if n=2,
        q=0.5 and a death is proposed (i.e. proposed value n*=1), then p_death=0.5 (equal probabilities of birth/death
        in CURRENT state), and p_birth=1 (because death is not possible in PROPOSED state).

        Args:
            current_state (dict): dictionary with current parameter values.
            birth (bool): indicator for birth or death move.

        Returns:
            p_birth (float): state-dependent probability of birth move.
            p_death (float): state-dependent probability of death move.

        """
        p_birth = self.birth_probability
        p_death = 1.0 - self.birth_probability

        if current_state[self.param] == self.n_max:
            p_death = 1.0
        if current_state[self.param] == (self.n_max - 1) and birth:
            p_death = 1.0

        if current_state[self.param] == 1:
            p_birth = 1.0
        if current_state[self.param] == 2 and not birth:
            p_birth = 1.0
        return p_birth, p_death



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Main MCMC class for mcmc setup."""

from copy import copy
from dataclasses import dataclass, field

import numpy as np
from scipy import sparse
from tqdm import tqdm

from openmcmc.model import Model
from openmcmc.sampler.metropolis_hastings import MetropolisHastings
from openmcmc.sampler.sampler import MCMCSampler


@dataclass
class MCMC:
    """Class for running Markov Chain Monte Carlo on a Model object to do parameter inference.

    Args:
        state (dict): initial state of sampler any parameters not
                     specified will be sampler from prior distributions
        samplers (list): list of the samplers to be used for each parameter to be estimated
        n_burn (int, optional): number of initial burn in these iterations are not stored, default 5000
        n_iter (int, optional): number of iterations which are stored in store, default 5000

    Attributes:
        state (dict): initial state of sampler any parameters not
                     specified will be sampler from prior distributions
        samplers (list): list of the samplers to be used for each parameter to be estimated.
        n_burn (int): number of initial burn in these iterations are not stored.
        n_iter (int): number of iterations which are stored in store.
        store (dict): dictionary storing MCMC output as np.array for each inference parameter.

    """

    state: dict
    samplers: list[MCMCSampler]
    model: Model
    n_burn: int = 5000
    n_iter: int = 5000
    store: dict = field(default_factory=dict, init=False)

    def __post_init__(self):
        """Convert any state values to at least 2D np.arrays and sample any missing states from the prior distributions, and set up storage arrays for the sampled values.

        Ensures that all elements of the initial state are in an appropriate format for running
        the sampler:
            - sparse matrices are left unchanged.
            - all other data types are coerced (if possible) to np.ndarray.
            - any scalars or existing np.ndarray with only one dimension are forced to be at
                least 2D.

        Also initialises an item in the storage dictionary for each of the sampled values,
        for any data fitted values, and for the log-posterior value.

        """
        self.state = copy(self.state)

        for key, term in self.state.items():
            if sparse.issparse(term):
                continue

            if not isinstance(term, np.ndarray):
                term = np.array(term, ndmin=2, dtype=np.float64)
                if np.shape(term)[0] == 1:
                    term = term.T
            elif term.ndim < 2:
                term = np.atleast_2d(term).T

            self.state[key] = term

        for sampler in self.samplers:
            if sampler.param not in self.state:
                self.state[sampler.param] = sampler.model[sampler.param].rvs(self.state)
            self.store = sampler.init_store(current_state=self.state, store=self.store, n_iterations=self.n_iter)
        if self.model.response is not None:
            for response in self.model.response.keys():
                self.store[response] = np.full(shape=(self.state[response].size, self.n_iter), fill_value=np.nan)
        self.store["log_post"] = np.full(shape=(self.n_iter, 1), fill_value=np.nan)

    def run_mcmc(self):
        """Runs MCMC routine for model specification loops for n_iter+ n_burn iterations sampling the state for each parameter and updating the parameter state.

        Runs a first loop over samplers, and generates a sample for all corresponding variables in the state. Then
        stores the value of each of the sampled parameters in the self.store dictionary, as well as the data fitted
        values and the log-posterior value.

        """
        for i_it in tqdm(range(-self.n_burn, self.n_iter)):
            for sampler in self.samplers:
                self.state = sampler.sample(self.state)

            if i_it < 0:
                continue

            for sampler in self.samplers:
                self.store = sampler.store(current_state=self.state, store=self.store, iteration=i_it)

            self.store["log_post"][i_it] = self.model.log_p(self.state)
            if self.model.response is not None:
                for response, predictor in self.model.response.items():
                    self.store[response][:, [i_it]] = getattr(self.model[response], predictor).predictor(self.state)

        for sampler in self.samplers:
            if isinstance(sampler, MetropolisHastings):
                print(f"{sampler.param}: {sampler.accept_rate.get_acceptance_rate()}")



# SPDX-FileCopyrightText: 2024 Shell Global Solutions International B.V. All Rights Reserved.
#
# SPDX-License-Identifier: Apache-2.0

"""Sampler module."""

__all__ = [
    "metropolis_hastings",
    "reversible_jump",
    "sampler",
]
